from typing import List, Optional, Dict, Tuple, Union

from chromadb import (
	EphemeralClient,
	PersistentClient,
	DEFAULT_TENANT,
	DEFAULT_DATABASE,
	CloudClient,
	AsyncHttpClient,
)
from chromadb.api.models.AsyncCollection import AsyncCollection
from chromadb.api.types import QueryResult

from autorag.utils.util import apply_recursive
from autorag.vectordb.base import BaseVectorStore


class Chroma(BaseVectorStore):
	def __init__(
		self,
		embedding_model: Union[str, List[dict]],
		collection_name: str,
		embedding_batch: int = 100,
		client_type: str = "persistent",
		similarity_metric: str = "cosine",
		path: str = None,
		host: str = "localhost",
		port: int = 8000,
		ssl: bool = False,
		headers: Optional[Dict[str, str]] = None,
		api_key: Optional[str] = None,
		tenant: str = DEFAULT_TENANT,
		database: str = DEFAULT_DATABASE,
	):
		super().__init__(embedding_model, similarity_metric, embedding_batch)
		if client_type == "ephemeral":
			self.client = EphemeralClient(tenant=tenant, database=database)
		elif client_type == "persistent":
			assert path is not None, "path must be provided for persistent client"
			self.client = PersistentClient(path=path, tenant=tenant, database=database)
		elif client_type == "http":
			self.client = AsyncHttpClient(
				host=host,
				port=port,
				ssl=ssl,
				headers=headers,
				tenant=tenant,
				database=database,
			)
		elif client_type == "cloud":
			self.client = CloudClient(
				tenant=tenant,
				database=database,
				api_key=api_key,
			)
		else:
			raise ValueError(
				f"client_type {client_type} is not supported\n"
				"supported client types are: ephemeral, persistent, http, cloud"
			)

		self.collection = self.client.get_or_create_collection(
			name=collection_name,
			metadata={"hnsw:space": similarity_metric},
		)

	async def add(self, ids: List[str], texts: List[str]):
		texts = self.truncated_inputs(texts)
		text_embeddings = await self.embedding.aget_text_embedding_batch(texts)
		if isinstance(self.collection, AsyncCollection):
			await self.collection.add(ids=ids, embeddings=text_embeddings)
		else:
			self.collection.add(ids=ids, embeddings=text_embeddings)

	def add_embedding(self, ids: List[str], embeddings: List[List[float]]):
		self.collection.add(ids=ids, embeddings=embeddings)

	async def fetch(self, ids: List[str]) -> List[List[float]]:
		if isinstance(self.collection, AsyncCollection):
			fetch_result = await self.collection.get(ids, include=["embeddings"])
		else:
			fetch_result = self.collection.get(ids, include=["embeddings"])
		fetch_embeddings = fetch_result["embeddings"]
		return fetch_embeddings

	async def is_exist(self, ids: List[str]) -> List[bool]:
		if isinstance(self.collection, AsyncCollection):
			fetched_result = await self.collection.get(ids, include=[])
		else:
			fetched_result = self.collection.get(ids, include=[])
		existed_ids = fetched_result["ids"]
		return list(map(lambda x: x in existed_ids, ids))

	async def query(
		self, queries: List[str], top_k: int, **kwargs
	) -> Tuple[List[List[str]], List[List[float]]]:
		queries = self.truncated_inputs(queries)
		query_embeddings: List[
			List[float]
		] = await self.embedding.aget_text_embedding_batch(queries)
		if isinstance(self.collection, AsyncCollection):
			query_result: QueryResult = await self.collection.query(
				query_embeddings=query_embeddings, n_results=top_k
			)
		else:
			query_result: QueryResult = self.collection.query(
				query_embeddings=query_embeddings, n_results=top_k
			)
		ids = query_result["ids"]
		scores = query_result["distances"]
		scores = apply_recursive(lambda x: 1 - x, scores)
		return ids, scores

	async def delete(self, ids: List[str]):
		if isinstance(self.collection, AsyncCollection):
			await self.collection.delete(ids)
		else:
			self.collection.delete(ids)



import logging
from typing import Callable, Optional, Dict, Awaitable, Any, Tuple, List
import uuid
import pandas as pd
from autorag.utils.util import process_batch, get_event_loop, fetch_contents

from autorag.support import get_support_modules

logger = logging.getLogger("AutoRAG")


class Raw:
	"""
	The Raw class that stored document parsing results.
	It can do chunking.
	It has two column names, 'raw_id' and 'contents'.
	"""

	def __init__(self, raw_df: Optional[pd.DataFrame] = None):
		self.data = raw_df

	def batch_apply(
		self, fn: Callable[[Dict, Any], Awaitable[Dict]], batch_size: int = 32, **kwargs
	) -> "Raw":
		raw_dicts = self.data.to_dict(orient="records")
		loop = get_event_loop()
		tasks = [fn(raw_dict, **kwargs) for raw_dict in raw_dicts]
		results = loop.run_until_complete(process_batch(tasks, batch_size))
		return Raw(pd.DataFrame(results))

	def map(self, fn: Callable[[pd.DataFrame, Any], pd.DataFrame], **kwargs) -> "Raw":
		return Raw(fn(self.data, **kwargs))

	def flatmap(self, fn: Callable, **kwargs) -> "Raw":
		return fn(self.data, **kwargs)

	def chunk(self, module_name: str, **module_params) -> "Corpus":
		chunk_module = get_support_modules(module_name)
		chunked_result = chunk_module(parsed_result=self.data, **module_params)
		return Corpus(chunked_result, self)

	def __add__(self, other):
		assert isinstance(other, Raw), "You can only add Raw instances."
		self.data = pd.concat([self.data, other.data], ignore_index=True).reset_index(
			drop=True
		)
		return self


class Corpus:
	"""
	The Corpus class that stored chunked passages.
	It can generate qa set, linked with Raw instance.
	"""

	def __init__(
		self,
		corpus_df: Optional[pd.DataFrame] = None,
		linked_raw: Optional[Raw] = None,
	):
		self.data = corpus_df
		self._linked_raw = linked_raw

	@property
	def linked_raw(self) -> Raw:
		return self._linked_raw

	@linked_raw.setter
	def linked_raw(self, raw: Raw):
		raise NotImplementedError("linked_raw is read-only.")

	def to_parquet(self, save_path: str):
		"""
		Save the corpus to the AutoRAG compatible parquet file.
		It is not for the data creation, for running AutoRAG.
		If you want to save it directly, use the below code.
		`corpus.data.to_parquet(save_path)`

		:param save_path: The path to save the corpus.
		"""
		if not save_path.endswith(".parquet"):
			raise ValueError("save_path must be ended with .parquet")
		save_df = self.data.reset_index(drop=True)
		save_df.to_parquet(save_path)

	def batch_apply(
		self, fn: Callable[[Dict, Any], Awaitable[Dict]], batch_size: int = 32, **kwargs
	) -> "Corpus":
		corpus_dicts = self.data.to_dict(orient="records")
		loop = get_event_loop()
		tasks = [fn(corpus_dict, **kwargs) for corpus_dict in corpus_dicts]
		results = loop.run_until_complete(process_batch(tasks, batch_size))
		return Corpus(pd.DataFrame(results), self.linked_raw)

	def map(
		self, fn: Callable[[pd.DataFrame, Any], pd.DataFrame], **kwargs
	) -> "Corpus":
		return Corpus(fn(self.data, **kwargs), self.linked_raw)

	def sample(self, fn: Callable[[pd.DataFrame, Any], pd.DataFrame], **kwargs) -> "QA":
		"""
		Sample the corpus for making QA.
		It selects the subset of the corpus and makes QA set from it.
		You can generate questions from the created question.
		It is the first step to make QA set from the corpus.
		If you select just one passage from each passage, it will be a single-hop QA set.
		If you select multiple passages from each passage, it will be a multi-hop QA set.

		:param fn: The select function to perform.
		It returns QA dataframe.
		:return: QA instance that is selected.
		It contains qid and retrieval_gt columns.
		"""
		return QA(fn(self.data, **kwargs), self)


class QA:
	def __init__(
		self,
		qa_df: Optional[pd.DataFrame] = None,
		linked_corpus: Optional[Corpus] = None,
	):
		self.data = qa_df
		self._linked_corpus = linked_corpus

	@property
	def linked_corpus(self) -> Corpus:
		return self._linked_corpus

	@linked_corpus.setter
	def linked_corpus(self, corpus: Corpus):
		raise NotImplementedError("linked_corpus is read-only.")

	def batch_apply(
		self, fn: Callable[[Dict, Any], Awaitable[Dict]], batch_size: int = 32, **kwargs
	) -> "QA":
		qa_dicts = self.data.to_dict(orient="records")
		loop = get_event_loop()
		tasks = [fn(qa_dict, **kwargs) for qa_dict in qa_dicts]
		results = loop.run_until_complete(process_batch(tasks, batch_size))

		# Experimental feature
		if fn.__name__ == "multiple_queries_gen":
			return self._process_multiple_queries_gen(results)

		return QA(pd.DataFrame(results), self.linked_corpus)

	def batch_filter(
		self, fn: Callable[[Dict, Any], Awaitable[bool]], batch_size: int = 32, **kwargs
	) -> "QA":
		qa_dicts = self.data.to_dict(orient="records")
		loop = get_event_loop()
		tasks = [fn(qa_dict, **kwargs) for qa_dict in qa_dicts]
		masks = loop.run_until_complete(process_batch(tasks, batch_size))
		return QA(self.data[masks], self.linked_corpus)

	def filter(self, fn: Callable[[Dict, Any], bool], **kwargs) -> "QA":
		qa_dicts = self.data.to_dict(orient="records")
		masks = [fn(qa_dict, **kwargs) for qa_dict in qa_dicts]
		return QA(self.data[masks], self.linked_corpus)

	def map(self, fn: Callable[[pd.DataFrame, Any], pd.DataFrame], **kwargs) -> "QA":
		return QA(fn(self.data, **kwargs), self.linked_corpus)

	def make_retrieval_gt_contents(self) -> "QA":
		"""
		Make retrieval_gt_contents column from retrieval_gt column.
		:return: The QA instance that has a retrieval_gt_contents column.
		"""
		self.data["retrieval_gt_contents"] = self.data["retrieval_gt"].apply(
			lambda x: fetch_contents(self.linked_corpus.data, x)
		)
		return self

	def to_parquet(self, qa_save_path: str, corpus_save_path: str):
		"""
		Save the qa and corpus to the AutoRAG compatible parquet file.
		It is not for the data creation, for running AutoRAG.
		If you want to save it directly, use the below code.
		`qa.data.to_parquet(save_path)`

		:param qa_save_path: The path to save the qa dataset.
		:param corpus_save_path: The path to save the corpus.
		"""
		if not qa_save_path.endswith(".parquet"):
			raise ValueError("save_path must be ended with .parquet")
		if not corpus_save_path.endswith(".parquet"):
			raise ValueError("save_path must be ended with .parquet")
		save_df = self.data[
			["qid", "query", "retrieval_gt", "generation_gt"]
		].reset_index(drop=True)
		save_df.to_parquet(qa_save_path)
		self.linked_corpus.to_parquet(corpus_save_path)

	def update_corpus(self, new_corpus: Corpus) -> "QA":
		"""
		Update linked corpus.
		Not just replace linked_corpus to the new Corpus,
		it replaces the whole `retrieval_gt` to the new corpus using `linked_raw`.
		The QA data must have a `retrieval_gt` column.

		:param new_corpus: Corpus that you want to replace.
		    Must have valid `linked_raw` and `raw_id`, `raw_start_idx`, `raw_end_idx` columns.
		:return: The QA instance that updated linked corpus.
		"""
		self.data["evidence_path"] = (
			self.data["retrieval_gt"]
			.apply(
				lambda x: fetch_contents(
					self.linked_corpus.data,
					x,
					column_name="path",
				)
			)
			.tolist()
		)
		self.data["evidence_page"] = self.data["retrieval_gt"].apply(
			lambda x: list(
				map(
					lambda lst: list(map(lambda x: x.get("page", -1), lst)),
					fetch_contents(self.linked_corpus.data, x, column_name="metadata"),
				)
			)
		)
		if "evidence_start_end_idx" not in self.data.columns:
			# make evidence start_end_idx
			self.data["evidence_start_end_idx"] = (
				self.data["retrieval_gt"]
				.apply(
					lambda x: fetch_contents(
						self.linked_corpus.data,
						x,
						column_name="start_end_idx",
					)
				)
				.tolist()
			)

		# matching the new corpus with the old corpus
		path_corpus_dict = QA.__make_path_corpus_dict(new_corpus.data)
		new_retrieval_gt = self.data.apply(
			lambda row: QA.__match_index_row(
				row["evidence_start_end_idx"],
				row["evidence_path"],
				row["evidence_page"],
				path_corpus_dict,
			),
			axis=1,
		).tolist()
		new_qa = self.data.copy(deep=True)[["qid", "query", "generation_gt"]]
		new_qa["retrieval_gt"] = new_retrieval_gt
		return QA(new_qa, new_corpus)

	@staticmethod
	def __match_index(target_idx: Tuple[int, int], dst_idx: Tuple[int, int]) -> bool:
		"""
		Check if the target_idx is overlap by the dst_idx.
		"""
		target_start, target_end = target_idx
		dst_start, dst_end = dst_idx
		return (
			dst_start <= target_start <= dst_end or dst_start <= target_end <= dst_end
		)

	@staticmethod
	def __match_index_row(
		evidence_indices: List[List[Tuple[int, int]]],
		evidence_paths: List[List[str]],
		evidence_pages: List[List[int]],
		path_corpus_dict: Dict,
	) -> List[List[str]]:
		"""
		Find the matched passage from new_corpus.

		:param evidence_indices: The evidence indices at the corresponding Raw.
		        Its shape is the same as the retrieval_gt.
		:param evidence_paths: The evidence paths at the corresponding Raw.
		        Its shape is the same as the retrieval_gt.
		:param path_corpus_dict: The key is the path name, and the value is the corpus dataframe that only contains the path in the key.
		        You can make it using `QA.__make_path_corpus_dict`.
		:return:
		"""
		result = []
		for i, idx_list in enumerate(evidence_indices):
			sub_result = []
			for j, idx in enumerate(idx_list):
				path_corpus_df = path_corpus_dict[evidence_paths[i][j]]
				if evidence_pages[i][j] >= 0:
					path_corpus_df = path_corpus_df.loc[
						path_corpus_df["metadata"].apply(lambda x: x.get("page", -1))
						== evidence_pages[i][j]
					]
				matched_corpus = path_corpus_df.loc[
					path_corpus_df["start_end_idx"].apply(
						lambda x: QA.__match_index(idx, x)
					)
				]
				sub_result.extend(matched_corpus["doc_id"].tolist())
			result.append(sub_result)
		return result

	@staticmethod
	def __make_path_corpus_dict(corpus_df: pd.DataFrame) -> Dict[str, pd.DataFrame]:
		return {
			path: corpus_df[corpus_df["path"] == path]
			for path in corpus_df["path"].unique()
		}

	# Experimental feature
	def _process_multiple_queries_gen(self, results: List[Dict]) -> "QA":
		data = []
		for result in results:
			queries = result["query"].split("\n")
			for query in queries:
				new_result = {
					key: (str(uuid.uuid4()) if key == "qid" else result[key])
					for key in result.keys()
				}
				new_result["query"] = query
				data.append(new_result)
		df = pd.DataFrame(data)
		return QA(df, self.linked_corpus)



import os
from typing import List, Callable, Dict
import pandas as pd
from glob import glob

from autorag.strategy import measure_speed
from autorag.data.utils.util import get_param_combinations

default_map = {
	"pdf": {
		"file_type": "pdf",
		"module_type": "langchain_parse",
		"parse_method": "pdfminer",
	},
	"csv": {
		"file_type": "csv",
		"module_type": "langchain_parse",
		"parse_method": "csv",
	},
	"md": {
		"file_type": "md",
		"module_type": "langchain_parse",
		"parse_method": "unstructuredmarkdown",
	},
	"html": {
		"file_type": "html",
		"module_type": "langchain_parse",
		"parse_method": "bshtml",
	},
	"xml": {
		"file_type": "xml",
		"module_type": "langchain_parse",
		"parse_method": "unstructuredxml",
	},
}


def run_parser(
	modules: List[Callable],
	module_params: List[Dict],
	data_path_glob: str,
	project_dir: str,
	all_files: bool,
):
	if not all_files:
		# Set the parsing module to default if it is a file type in paths but not set in YAML.
		data_path_list = glob(data_path_glob)
		if not data_path_list:
			raise FileNotFoundError(f"data does not exits in {data_path_glob}")

		file_types = set(
			[os.path.basename(data_path).split(".")[-1] for data_path in data_path_list]
		)
		set_file_types = set([module["file_type"] for module in module_params])

		# Calculate the set difference once
		file_types_to_remove = set_file_types - file_types

		# Use list comprehension to filter out unwanted elements
		module_params = [
			param
			for param in module_params
			if param["file_type"] not in file_types_to_remove
		]
		modules = [
			module
			for module, param in zip(modules, module_params)
			if param["file_type"] not in file_types_to_remove
		]

		# create a list of only those file_types that are in file_types but not in set_file_types
		missing_file_types = list(file_types - set_file_types)

		if missing_file_types:
			add_modules_list = []
			for missing_file_type in missing_file_types:
				if missing_file_type == "json":
					raise ValueError(
						"JSON file type must have a jq_schema so you must set it in the YAML file."
					)

				add_modules_list.append(default_map[missing_file_type])

			add_modules, add_params = get_param_combinations(add_modules_list)
			modules.extend(add_modules)
			module_params.extend(add_params)

	results, execution_times = zip(
		*map(
			lambda x: measure_speed(x[0], data_path_glob=data_path_glob, **x[1]),
			zip(modules, module_params),
		)
	)
	average_times = list(map(lambda x: x / len(results[0]), execution_times))

	# save results to parquet files
	if all_files:
		if len(module_params) > 1:
			raise ValueError(
				"All files is set to True, You can only use one parsing module."
			)
		filepaths = [os.path.join(project_dir, "parsed_result.parquet")]
	else:
		filepaths = list(
			map(
				lambda x: os.path.join(project_dir, f"{x['file_type']}.parquet"),
				module_params,
			)
		)

	_files = {}
	for result, filepath in zip(results, filepaths):
		_files[filepath].append(result) if filepath in _files.keys() else _files.update(
			{filepath: [result]}
		)
	# Save files with a specific file type as Parquet files.
	for filepath, value in _files.items():
		pd.concat(value).to_parquet(filepath, index=False)

	filenames = list(map(lambda x: os.path.basename(x), filepaths))

	summary_df = pd.DataFrame(
		{
			"filename": filenames,
			"module_name": list(map(lambda module: module.__name__, modules)),
			"module_params": module_params,
			"execution_time": average_times,
		}
	)
	summary_df.to_csv(os.path.join(project_dir, "summary.csv"), index=False)

	# concat all parquet files here if not all_files.
	_filepaths = list(_files.keys())
	if not all_files:
		dataframes = [pd.read_parquet(file) for file in _filepaths]
		combined_df = pd.concat(dataframes, ignore_index=True)
		combined_df.to_parquet(
			os.path.join(project_dir, "parsed_result.parquet"), index=False
		)

	return summary_df



import functools
import warnings
from typing import List, Callable, Any, Tuple, Union, Dict

import pandas as pd

from autorag.evaluation.metric import (
	retrieval_recall,
	retrieval_precision,
	retrieval_f1,
	retrieval_ndcg,
	retrieval_mrr,
	retrieval_map,
)
from autorag.evaluation.util import cast_metrics
from autorag.schema.metricinput import MetricInput

RETRIEVAL_METRIC_FUNC_DICT = {
	func.__name__: func
	for func in [
		retrieval_recall,
		retrieval_precision,
		retrieval_f1,
		retrieval_ndcg,
		retrieval_mrr,
		retrieval_map,
	]
}


def evaluate_retrieval(
	metric_inputs: List[MetricInput],
	metrics: Union[List[str], List[Dict]],
):
	def decorator_evaluate_retrieval(
		func: Callable[
			[Any], Tuple[List[List[str]], List[List[str]], List[List[float]]]
		],
	):
		"""
		Decorator for evaluating retrieval results.
		You can use this decorator to any method that returns (contents, scores, ids),
		which is the output of conventional retrieval modules.

		:param func: Must return (contents, scores, ids)
		:return: wrapper function that returns pd.DataFrame, which is the evaluation result.
		"""

		@functools.wraps(func)
		def wrapper(*args, **kwargs) -> pd.DataFrame:
			contents, pred_ids, scores = func(*args, **kwargs)
			for metric_input, pred_id in zip(metric_inputs, pred_ids):
				metric_input.retrieved_ids = pred_id

			metric_scores = {}
			metric_names, metric_params = cast_metrics(metrics)

			for metric_name, metric_param in zip(metric_names, metric_params):
				if metric_name in RETRIEVAL_METRIC_FUNC_DICT:
					metric_func = RETRIEVAL_METRIC_FUNC_DICT[metric_name]
					metric_scores[metric_name] = metric_func(
						metric_inputs=metric_inputs, **metric_param
					)
				else:
					warnings.warn(
						f"metric {metric_name} is not in supported metrics: {RETRIEVAL_METRIC_FUNC_DICT.keys()}"
						f"{metric_name} will be ignored."
					)

			metric_result_df = pd.DataFrame(metric_scores)
			execution_result_df = pd.DataFrame(
				{
					"retrieved_contents": contents,
					"retrieved_ids": pred_ids,
					"retrieve_scores": scores,
				}
			)
			result_df = pd.concat([execution_result_df, metric_result_df], axis=1)
			return result_df

		return wrapper

	return decorator_evaluate_retrieval



import importlib.resources
import logging
import os
import pathlib
import subprocess
from pathlib import Path
from typing import Optional

import click
import nest_asyncio

from autorag import dashboard
from autorag.deploy import extract_best_config as original_extract_best_config
from autorag.deploy.api import ApiRunner
from autorag.evaluator import Evaluator
from autorag.validator import Validator

logger = logging.getLogger("AutoRAG")

autorag_dir = os.path.dirname(os.path.realpath(__file__))
version_file = os.path.join(autorag_dir, "VERSION")
with open(version_file, "r") as f:
	__version__ = f.read().strip()


@click.group()
@click.version_option(__version__)
def cli():
	pass


@click.command()
@click.option(
	"--config",
	"-c",
	help="Path to config yaml file. Must be yaml or yml file.",
	type=str,
)
@click.option(
	"--qa_data_path", help="Path to QA dataset. Must be parquet file.", type=str
)
@click.option(
	"--corpus_data_path", help="Path to corpus dataset. Must be parquet file.", type=str
)
@click.option(
	"--project_dir", help="Path to project directory.", type=str, default=None
)
@click.option(
	"--skip_validation",
	help="Skip validation or not. Default is False.",
	type=bool,
	default=False,
)
def evaluate(config, qa_data_path, corpus_data_path, project_dir, skip_validation):
	if not config.endswith(".yaml") and not config.endswith(".yml"):
		raise ValueError(f"Config file {config} is not a yaml or yml file.")
	if not os.path.exists(config):
		raise ValueError(f"Config file {config} does not exist.")
	evaluator = Evaluator(qa_data_path, corpus_data_path, project_dir=project_dir)
	evaluator.start_trial(config, skip_validation=skip_validation)


@click.command()
@click.option(
	"--config_path", type=str, help="Path to extracted config yaml file.", default=None
)
@click.option("--host", type=str, default="0.0.0.0", help="Host address")
@click.option("--port", type=int, default=8000, help="Port number")
@click.option(
	"--trial_dir",
	type=click.Path(file_okay=False, dir_okay=True, exists=True),
	default=None,
	help="Path to trial directory.",
)
@click.option(
	"--project_dir", help="Path to project directory.", type=str, default=None
)
@click.option(
	"--remote", help="Run the API server in remote mode.", type=bool, default=False
)
def run_api(config_path, host, port, trial_dir, project_dir, remote: bool):
	if trial_dir is None:
		runner = ApiRunner.from_yaml(config_path, project_dir=project_dir)
	else:
		runner = ApiRunner.from_trial_folder(trial_dir)
	logger.info(f"Running API server at {host}:{port}...")
	nest_asyncio.apply()
	runner.run_api_server(host, port, remote=remote)


@click.command()
@click.option(
	"--yaml_path", type=click.Path(path_type=Path), help="Path to the YAML file."
)
@click.option(
	"--project_dir",
	type=click.Path(path_type=Path),
	help="Path to the project directory.",
)
@click.option(
	"--trial_path", type=click.Path(path_type=Path), help="Path to the trial directory."
)
def run_web(
	yaml_path: Optional[str], project_dir: Optional[str], trial_path: Optional[str]
):
	try:
		with importlib.resources.path("autorag", "web.py") as web_path:
			web_py_path = str(web_path)
	except ImportError:
		raise ImportError(
			"Could not locate the web.py file within the autorag package."
			" Please ensure that autorag is correctly installed."
		)

	if not yaml_path and not trial_path:
		raise ValueError("yaml_path or trial_path must be given.")
	elif yaml_path and trial_path:
		raise ValueError("yaml_path and trial_path cannot be given at the same time.")
	elif yaml_path and not project_dir:
		subprocess.run(
			["streamlit", "run", web_py_path, "--", "--yaml_path", yaml_path]
		)
	elif yaml_path and project_dir:
		subprocess.run(
			[
				"streamlit",
				"run",
				web_py_path,
				"--",
				"--yaml_path",
				yaml_path,
				"--project_dir",
				project_dir,
			]
		)
	elif trial_path:
		subprocess.run(
			["streamlit", "run", web_py_path, "--", "--trial_path", trial_path]
		)


@click.command()
@click.option(
	"--trial_dir",
	type=click.Path(dir_okay=True, file_okay=False, exists=True),
	required=True,
)
@click.option(
	"--port", type=int, default=7690, help="Port number. The default is 7690."
)
def run_dashboard(trial_dir: str, port: int):
	dashboard.run(trial_dir, port=port)


@click.command()
@click.option("--trial_path", type=click.Path(), help="Path to the trial directory.")
@click.option(
	"--output_path",
	type=click.Path(),
	help="Path to the output directory." " Must be .yaml or .yml file.",
)
def extract_best_config(trial_path: str, output_path: str):
	original_extract_best_config(trial_path, output_path)


@click.command()
@click.option("--trial_path", help="Path to trial directory.", type=str)
def restart_evaluate(trial_path):
	if not os.path.exists(trial_path):
		raise ValueError(f"trial_path {trial_path} does not exist.")
	project_dir = str(pathlib.PurePath(trial_path).parent)
	qa_data_path = os.path.join(project_dir, "data", "qa.parquet")
	corpus_data_path = os.path.join(project_dir, "data", "corpus.parquet")
	evaluator = Evaluator(qa_data_path, corpus_data_path, project_dir)
	evaluator.restart_trial(trial_path)


@click.command()
@click.option(
	"--config",
	"-c",
	help="Path to config yaml file. Must be yaml or yml file.",
	type=str,
)
@click.option(
	"--qa_data_path", help="Path to QA dataset. Must be parquet file.", type=str
)
@click.option(
	"--corpus_data_path", help="Path to corpus dataset. Must be parquet file.", type=str
)
def validate(config, qa_data_path, corpus_data_path):
	if not config.endswith(".yaml") and not config.endswith(".yml"):
		raise ValueError(f"Config file {config} is not a parquet file.")
	if not os.path.exists(config):
		raise ValueError(f"Config file {config} does not exist.")
	validator = Validator(qa_data_path=qa_data_path, corpus_data_path=corpus_data_path)
	validator.validate(config)


cli.add_command(evaluate, "evaluate")
cli.add_command(run_api, "run_api")
cli.add_command(run_web, "run_web")
cli.add_command(run_dashboard, "dashboard")
cli.add_command(extract_best_config, "extract_best_config")
cli.add_command(restart_evaluate, "restart_evaluate")
cli.add_command(validate, "validate")

if __name__ == "__main__":
	cli()



import logging

import weaviate
from weaviate.classes.init import Auth
from weaviate.classes.config import Property, DataType
import weaviate.classes as wvc
from weaviate.classes.query import MetadataQuery

from typing import List, Optional, Tuple, Union

from autorag.vectordb import BaseVectorStore

logger = logging.getLogger("AutoRAG")


class Weaviate(BaseVectorStore):
	def __init__(
		self,
		embedding_model: Union[str, List[dict]],
		collection_name: str,
		embedding_batch: int = 100,
		similarity_metric: str = "cosine",
		client_type: str = "docker",
		host: str = "localhost",
		port: int = 8080,
		grpc_port: int = 50051,
		url: Optional[str] = None,
		api_key: Optional[str] = None,
		text_key: str = "content",
	):
		super().__init__(embedding_model, similarity_metric, embedding_batch)

		self.text_key = text_key

		if client_type == "docker":
			self.client = weaviate.connect_to_local(
				host=host,
				port=port,
				grpc_port=grpc_port,
			)
		elif client_type == "cloud":
			self.client = weaviate.connect_to_weaviate_cloud(
				cluster_url=url,
				auth_credentials=Auth.api_key(api_key),
			)
		else:
			raise ValueError(
				f"client_type {client_type} is not supported\n"
				"supported client types are: docker, cloud"
			)
		if similarity_metric == "cosine":
			distance_metric = wvc.config.VectorDistances.COSINE
		elif similarity_metric == "ip":
			distance_metric = wvc.config.VectorDistances.DOT
		elif similarity_metric == "l2":
			distance_metric = wvc.config.VectorDistances.L2_SQUARED
		else:
			raise ValueError(
				f"similarity_metric {similarity_metric} is not supported\n"
				"supported similarity metrics are: cosine, ip, l2"
			)

		if not self.client.collections.exists(collection_name):
			self.client.collections.create(
				collection_name,
				properties=[
					Property(
						name="content", data_type=DataType.TEXT, skip_vectorization=True
					),
				],
				vectorizer_config=wvc.config.Configure.Vectorizer.none(),
				vector_index_config=wvc.config.Configure.VectorIndex.hnsw(  # hnsw, flat, dynamic,
					distance_metric=distance_metric
				),
			)
		self.collection = self.client.collections.get(collection_name)
		self.collection_name = collection_name

	async def add(self, ids: List[str], texts: List[str]):
		texts = self.truncated_inputs(texts)
		text_embeddings = await self.embedding.aget_text_embedding_batch(texts)

		with self.client.batch.dynamic() as batch:
			for i, text in enumerate(texts):
				data_properties = {self.text_key: text}

				batch.add_object(
					collection=self.collection_name,
					properties=data_properties,
					uuid=ids[i],
					vector=text_embeddings[i],
				)

		failed_objs = self.client.batch.failed_objects
		for obj in failed_objs:
			err_message = (
				f"Failed to add object: {obj.original_uuid}\nReason: {obj.message}"
			)

			logger.error(err_message)

	def add_embedding(self, ids: List[str], embeddings: List[List[float]]):
		with self.client.batch.dynamic() as batch:
			for i in range(len(ids)):
				batch.add_object(
					collection=self.collection_name,
					uuid=ids[i],
					vector=embeddings[i],
				)

		failed_objs = self.client.batch.failed_objects
		for obj in failed_objs:
			err_message = (
				f"Failed to add object: {obj.original_uuid}\nReason: {obj.message}"
			)

			logger.error(err_message)

	async def fetch(self, ids: List[str]) -> List[List[float]]:
		# Fetch vectors by IDs
		results = self.collection.query.fetch_objects(
			filters=wvc.query.Filter.by_property("_id").contains_any(ids),
			include_vector=True,
		)
		id_vector_dict = {
			str(object.uuid): object.vector["default"] for object in results.objects
		}
		result = [id_vector_dict[_id] for _id in ids]
		return result

	async def is_exist(self, ids: List[str]) -> List[bool]:
		fetched_result = self.collection.query.fetch_objects(
			filters=wvc.query.Filter.by_property("_id").contains_any(ids),
		)
		existed_ids = [str(result.uuid) for result in fetched_result.objects]
		return list(map(lambda x: x in existed_ids, ids))

	async def query(
		self, queries: List[str], top_k: int, **kwargs
	) -> Tuple[List[List[str]], List[List[float]]]:
		queries = self.truncated_inputs(queries)
		query_embeddings: List[
			List[float]
		] = await self.embedding.aget_text_embedding_batch(queries)

		ids, scores = [], []
		for query_embedding in query_embeddings:
			response = self.collection.query.near_vector(
				near_vector=query_embedding,
				limit=top_k,
				return_metadata=MetadataQuery(distance=True),
			)

			ids.append([o.uuid for o in response.objects])
			scores.append(
				[
					distance_to_score(o.metadata.distance, self.similarity_metric)
					for o in response.objects
				]
			)

		return ids, scores

	async def delete(self, ids: List[str]):
		filter = wvc.query.Filter.by_id().contains_any(ids)
		self.collection.data.delete_many(where=filter)

	def delete_collection(self):
		# Delete the collection
		self.client.collections.delete(self.collection_name)


def distance_to_score(distance: float, similarity_metric) -> float:
	if similarity_metric == "cosine":
		return 1 - distance
	elif similarity_metric == "ip":
		return -distance
	elif similarity_metric == "l2":
		return -distance
	else:
		raise ValueError(
			f"similarity_metric {similarity_metric} is not supported\n"
			"supported similarity metrics are: cosine, ip, l2"
		)



import functools
import time
from typing import List, Iterable, Tuple, Any, Optional, Callable

import numpy as np
import pandas as pd


def measure_speed(func, *args, **kwargs):
	"""
	Method for measuring execution speed of the function.
	"""
	start_time = time.time()
	result = func(*args, **kwargs)
	end_time = time.time()
	return result, end_time - start_time


def avoid_empty_result(return_index: List[int]):
	"""
	Decorator for avoiding empty results from the function.
	When the func returns an empty result, it will return the origin results.
	When the func returns a None, it will return the origin results.
	When the return value is a tuple, it will check all the value or list is empty.
	If so, it will return the origin results.
	It keeps parameters at return_index of the function as the origin results.

	:param return_index: The index of the result to be returned when there is no result.
	:return: The origin results or the results from the function.
	"""

	def decorator_avoid_empty_result(func: Callable):
		@functools.wraps(func)
		def wrapper(*args, **kwargs) -> List:
			func_result = func(*args, **kwargs)
			if isinstance(func_result, tuple):
				# if all the results are empty, return the origin results.
				if all([not bool(result) for result in func_result]):
					return [args[index] for index in return_index]
			if not bool(func_result):
				return [args[index] for index in return_index]
			else:
				return func_result

		return wrapper

	return decorator_avoid_empty_result


@avoid_empty_result([0, 3])
def filter_by_threshold(results, value, threshold, metadatas=None) -> Tuple[List, List]:
	"""
	Filter results by value's threshold.

	:param results: The result list to be filtered.
	:param value: The value list to be filtered.
	    It must have the same length with results.
	:param threshold: The threshold value.
	:param metadatas: The metadata of each result.
	:return: Filtered list of results and filtered list of metadatas.
	    Metadatas will be returned even if you did not give input metadatas.
	:rtype: Tuple[List, List]
	"""
	if metadatas is None:
		metadatas = [None] * len(results)
	assert len(results) == len(value), "results and value must have the same length."
	try:
		filtered_results, _, filtered_metadatas = zip(
			*filter(lambda x: x[1] <= threshold, zip(results, value, metadatas))
		)
	except ValueError:
		return [], []
	return list(filtered_results), list(filtered_metadatas)


def validate_strategy_inputs(
	results: List[pd.DataFrame],
	columns: Iterable[str],
	metadatas: Optional[List[Any]] = None,
):
	if metadatas is None:
		metadatas = [None] * len(results)
	assert len(results) == len(
		metadatas
	), "results and module_filename must have the same length."
	assert all(
		[isinstance(result, pd.DataFrame) for result in results]
	), "results must be pd.DataFrame."
	assert all(
		[column in result.columns for result in results for column in columns]
	), "columns must be in the columns of results."
	return results, columns, metadatas


def select_best(
	results: List[pd.DataFrame],
	columns: Iterable[str],
	metadatas: Optional[List[Any]] = None,
	strategy_name: str = "mean",
) -> Tuple[pd.DataFrame, Any]:
	strategy_func_dict = {
		"mean": select_best_average,
		"rank": select_best_rr,
		"normalize_mean": select_normalize_mean,
	}
	if strategy_name not in strategy_func_dict:
		raise ValueError(
			f"Input strategy name {strategy_name} is not in {strategy_func_dict.keys()}"
		)

	return strategy_func_dict[strategy_name](results, columns, metadatas)


def select_best_average(
	results: List[pd.DataFrame],
	columns: Iterable[str],
	metadatas: Optional[List[Any]] = None,
) -> Tuple[pd.DataFrame, Any]:
	"""
	Select the best result by average value among given columns.

	:param results: The list of results.
	    Each result must be pd.DataFrame.
	:param columns: Column names to be averaged.
	    Standard to select the best result.
	:param metadatas: The metadata of each result.
	    It will select one metadata with the best result.
	:return: The best result and the best metadata.
	    The metadata will be returned even if you did not give input 'metadatas' parameter.
	:rtype: Tuple[pd.DataFrame, Any]
	"""
	results, columns, metadatas = validate_strategy_inputs(results, columns, metadatas)
	each_average = [df[columns].mean(axis=1).mean() for df in results]
	best_index = each_average.index(max(each_average))
	return results[best_index], metadatas[best_index]


def select_best_rr(
	results: List[pd.DataFrame],
	columns: Iterable[str],
	metadatas: Optional[List[Any]] = None,
) -> Tuple[pd.DataFrame, Any]:
	results, columns, metadatas = validate_strategy_inputs(results, columns, metadatas)
	each_average_df = pd.DataFrame(
		[df[columns].mean(axis=0).to_dict() for df in results]
	)
	rank_df = each_average_df.rank(ascending=False)
	rr_df = rank_df.map(lambda x: 1 / x)
	best_index = np.array(rr_df.sum(axis=1)).argmax()
	return results[best_index], metadatas[best_index]


def select_normalize_mean(
	results: List[pd.DataFrame],
	columns: Iterable[str],
	metadatas: Optional[List[Any]] = None,
) -> Tuple[pd.DataFrame, Any]:
	results, columns, metadatas = validate_strategy_inputs(results, columns, metadatas)
	each_mean_df = pd.DataFrame([df[columns].mean(axis=0).to_dict() for df in results])
	normalized_means = (each_mean_df - each_mean_df.min()) / (
		each_mean_df.max() - each_mean_df.min()
	)
	normalized_mean_sums = normalized_means.sum(axis=1)
	best_index = normalized_mean_sums.argmax()
	return results[best_index], metadatas[best_index]



from typing import List, Union

import numpy as np
import pandas as pd

from autorag.embedding.base import EmbeddingModel
from autorag.evaluation.metric.util import calculate_cosine_similarity
from autorag.nodes.passageaugmenter.base import BasePassageAugmenter
from autorag.utils.util import (
	filter_dict_keys,
	fetch_contents,
	embedding_query_content,
	result_to_dataframe,
	empty_cuda_cache,
)


class PrevNextPassageAugmenter(BasePassageAugmenter):
	def __init__(
		self,
		project_dir: str,
		embedding_model: Union[str, dict] = "openai",
		*args,
		**kwargs,
	):
		"""
		Initialize the PrevNextPassageAugmenter module.

		:param project_dir:
		:param embedding_model: The embedding model name to use for calculating cosine similarity
			Default is openai (text-embedding-ada-002)
		:param kwargs:
		"""
		super().__init__(project_dir, *args, **kwargs)
		slim_corpus_df = self.corpus_df[["doc_id", "metadata"]]
		slim_corpus_df.loc[:, "metadata"] = slim_corpus_df["metadata"].apply(
			filter_dict_keys, keys=["prev_id", "next_id"]
		)
		self.slim_corpus_df = slim_corpus_df

		# init embedding model
		self.embedding_model = EmbeddingModel.load(embedding_model)()

	def __del__(self):
		del self.embedding_model
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		"""
		Run the passage augmenter node - PrevNextPassageAugmenter module.

		:param previous_result: The previous result Dataframe.
		:param top_k: You must input the top_k value to get the top k results.
		:param kwargs: Not affected.
		:return: DataFrame with retrieved_contents, retrieved_ids, and retrieve_scores columns
		"""
		top_k = kwargs.pop("top_k")

		ids = self.cast_to_run(previous_result)
		# find queries columns
		assert (
			"query" in previous_result.columns
		), "previous_result must have query column."
		queries = previous_result["query"].tolist()

		mode = kwargs.pop("mode", "both")
		num_passages = kwargs.pop("num_passages", 1)
		augmented_ids = self._pure(ids, num_passages, mode)

		# fetch contents from corpus to use augmented ids
		augmented_contents = fetch_contents(self.corpus_df, augmented_ids)

		query_embeddings, contents_embeddings = embedding_query_content(
			queries, augmented_contents, self.embedding_model, batch=128
		)

		# get scores from calculated cosine similarity
		augmented_scores = [
			np.array(
				[
					calculate_cosine_similarity(query_embedding, x)
					for x in content_embeddings
				]
			).tolist()
			for query_embedding, content_embeddings in zip(
				query_embeddings, contents_embeddings
			)
		]
		return self.sort_by_scores(
			augmented_contents, augmented_ids, augmented_scores, top_k
		)

	def _pure(
		self,
		ids_list: List[List[str]],
		num_passages: int = 1,
		mode: str = "both",
	) -> List[List[str]]:
		"""
		Add passages before and/or after the retrieved passage.
		For more information, visit https://docs.llamaindex.ai/en/stable/examples/node_postprocessor/PrevNextPostprocessorDemo/.

		:param ids_list: The list of lists of ids retrieved
		:param num_passages: The number of passages to add before and after the retrieved passage
		    Default is 1.
		:param mode: The mode of augmentation
		    'prev': add passages before the retrieved passage
		    'next': add passages after the retrieved passage
		    'both': add passages before and after the retrieved passage
		    Default is 'next'.
		:return: The list of lists of augmented ids
		"""
		if mode not in ["prev", "next", "both"]:
			raise ValueError(f"mode must be 'prev', 'next', or 'both', but got {mode}")

		augmented_ids = [
			(
				lambda ids: prev_next_augmenter_pure(
					ids, self.slim_corpus_df, mode, num_passages
				)
			)(ids)
			for ids in ids_list
		]

		return augmented_ids


def prev_next_augmenter_pure(
	ids: List[str], corpus_df: pd.DataFrame, mode: str, num_passages: int
):
	def fetch_id_sequence(start_id, key):
		sequence = []
		current_id = start_id
		for _ in range(num_passages):
			current_id = (
				corpus_df.loc[corpus_df["doc_id"] == current_id]["metadata"]
				.values[0]
				.get(key)
			)
			if current_id is None:
				break
			sequence.append(current_id)
		return sequence

	augmented_group = []
	for id_ in ids:
		current_ids = [id_]
		if mode in ["prev", "both"]:
			current_ids = fetch_id_sequence(id_, "prev_id")[::-1] + current_ids
		if mode in ["next", "both"]:
			current_ids += fetch_id_sequence(id_, "next_id")
		augmented_group.extend(current_ids)
	return augmented_group



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from typing import Any, Dict, List, Optional

from transformers import T5Tokenizer


class EncT5Tokenizer(T5Tokenizer):
	def __init__(
		self,
		vocab_file,
		bos_token="<s>",
		eos_token="</s>",
		unk_token="<unk>",
		pad_token="<pad>",
		extra_ids=100,
		additional_special_tokens=None,
		sp_model_kwargs: Optional[Dict[str, Any]] = None,
		**kwargs,
	) -> None:
		sp_model_kwargs = {} if sp_model_kwargs is None else sp_model_kwargs

		super().__init__(
			vocab_file=vocab_file,
			bos_token=bos_token,
			eos_token=eos_token,
			unk_token=unk_token,
			pad_token=pad_token,
			extra_ids=extra_ids,
			additional_special_tokens=additional_special_tokens,
			sp_model_kwargs=sp_model_kwargs,
			**kwargs,
		)

	def get_special_tokens_mask(
		self,
		token_ids_0: List[int],
		token_ids_1: Optional[List[int]] = None,
		already_has_special_tokens: bool = False,
	) -> List[int]:
		"""
		Retrieve sequence ids from a token list that has no special tokens added. This method is called when adding
		special tokens using the tokenizer `prepare_for_model` method.
		Args:
		    token_ids_0 (`List[int]`):
		        List of IDs.
		    token_ids_1 (`List[int]`, *optional*):
		        Optional second list of IDs for sequence pairs.
		    already_has_special_tokens (`bool`, *optional*, defaults to `False`):
		        Whether or not the token list is already formatted with special tokens for the model.
		Returns:
		    `List[int]`: A list of integers in the range [0, 1]: 1 for a special token, 0 for a sequence token.
		"""
		if already_has_special_tokens:
			return super().get_special_tokens_mask(
				token_ids_0=token_ids_0,
				token_ids_1=token_ids_1,
				already_has_special_tokens=True,
			)

		# normal case: some special tokens
		if token_ids_1 is None:
			return [1] + ([0] * len(token_ids_0)) + [1]
		return [1] + ([0] * len(token_ids_0)) + [1] + ([0] * len(token_ids_1)) + [1]

	def create_token_type_ids_from_sequences(
		self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
	) -> List[int]:
		"""
		Create a mask from the two sequences passed to be used in a sequence-pair classification task. T5 does not make
		use of token type ids, therefore a list of zeros is returned.
		Args:
		    token_ids_0 (`List[int]`):
		        List of IDs.
		    token_ids_1 (`List[int]`, *optional*):
		        Optional second list of IDs for sequence pairs.
		Returns:
		    `List[int]`: List of zeros.
		"""
		bos = [self.bos_token_id]
		eos = [self.eos_token_id]

		if token_ids_1 is None:
			return len(bos + token_ids_0 + eos) * [0]
		return len(bos + token_ids_0 + eos + token_ids_1 + eos) * [0]

	def build_inputs_with_special_tokens(
		self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
	) -> List[int]:
		"""
		Build model inputs from a sequence or a pair of sequence for sequence classification tasks by concatenating and
		adding special tokens. A sequence has the following format:
		- single sequence: `<s> X </s>`
		- pair of sequences: `<s> A </s> B </s>`
		Args:
		    token_ids_0 (`List[int]`):
		        List of IDs to which the special tokens will be added.
		    token_ids_1 (`List[int]`, *optional*):
		        Optional second list of IDs for sequence pairs.
		Returns:
		    `List[int]`: List of [input IDs](../glossary#input-ids) with the appropriate special tokens.
		"""
		if token_ids_1 is None:
			return [self.bos_token_id] + token_ids_0 + [self.eos_token_id]
		else:
			return (
				[self.bos_token_id]
				+ token_ids_0
				+ [self.eos_token_id]
				+ token_ids_1
				+ [self.eos_token_id]
			)



from typing import Dict, List

from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.base.llms.types import ChatMessage, MessageRole, ChatResponse
from llama_index.llms.openai.utils import to_openai_message_dicts
from openai import AsyncClient
from pydantic import BaseModel

from autorag.data.qa.filter.prompt import FILTER_PROMPT

dont_know_phrases = {
	"en": [
		"I don't know",
		"I do not know",
		"Don't know",
		"Do not know",
	],
	"ko": [
		"몰라요",
		"모르겠습니다",
		"모르겠어요",
		"몰라",
		"내가 어떻게 알아?",
		"모르겠소",
		"몰라유",
		"모르것는디",
		"모르겠어유",
		"모르겠네유",
		"모르겠네요",
	],
	"ja": [
		"知りません",
		"わかりません",
		"分かりません",
		"知らないです",
		"よく分かってません",
		"わかりかねます",
		"存じません",
		"お答えいたしかねます",
	],
}


def dontknow_filter_rule_based(row: Dict, lang: str = "en") -> bool:
	assert (
		"generation_gt" in row.keys()
	), "generation_gt column is not in the DataFrame."
	dont_know_phrase = dont_know_phrases[lang]
	return not any(
		phrase in s for phrase in dont_know_phrase for s in row["generation_gt"]
	)


class Response(BaseModel):
	is_dont_know: bool


async def dontknow_filter_openai(
	row: Dict,
	client: AsyncClient,
	model_name: str = "gpt-4o-mini-2024-07-18",
	lang: str = "en",
) -> bool:
	"""
	This will drop rows that have a "don't know" answer.
	It will drop unanswerable questions from the QA dataset.
	You can use this filter with the ` batch_filter ` function at `QA` class.

	:param row: The row dict from QA dataset.
	:param client: The OpenAI client.
	:param model_name: The model name.
		You have to use gpt-4o-2024-08-06 or gpt-4o-mini-2024-07-18.
	:param lang: The supported language is en, ko or ja.
	:return: False if the row generation_gt is a "don't know" meaning.
	"""
	assert "generation_gt" in row.keys(), "generation_gt column is not in the row."
	system_prompt: List[ChatMessage] = FILTER_PROMPT["dontknow_filter"][lang]
	result = []
	for gen_gt in row["generation_gt"]:
		completion = await client.beta.chat.completions.parse(
			model=model_name,
			messages=to_openai_message_dicts(
				system_prompt + [ChatMessage(role=MessageRole.USER, content=gen_gt)]
			),
			response_format=Response,
		)
		result.append(completion.choices[0].message.parsed.is_dont_know)
	return not any(result)


async def dontknow_filter_llama_index(
	row: Dict,
	llm: BaseLLM,
	lang: str = "en",
) -> bool:
	"""
	This will drop rows that have a "don't know" answer.
	It will drop unanswerable questions from the QA dataset.
	You can use this filter with the ` batch_filter ` function at `QA` class.

	:param row: The row dict from QA dataset.
	:param llm: The Llama index llm instance.
		It will be good if you set max tokens to low for saving tokens.
	:param lang: The supported language is en, ko or ja.
	:return: False if the row generation_gt is a "don't know" meaning.
	"""
	assert "generation_gt" in row.keys(), "generation_gt column is not in the row."
	system_prompt: List[ChatMessage] = FILTER_PROMPT["dontknow_filter"][lang]
	results = []
	for gen_gt in row["generation_gt"]:
		response: ChatResponse = await llm.achat(
			messages=system_prompt
			+ [ChatMessage(role=MessageRole.USER, content=gen_gt)]
		)
		result_str = response.message.content
		results.append("true" in result_str.lower().strip())
	return not any(results)



from typing import List, Optional

from llama_index.core import PromptTemplate
from llama_index.core.prompts import PromptType
from llama_index.core.prompts.utils import is_chat_model
from llama_index.core.response_synthesizers import TreeSummarize as ts

from autorag.nodes.passagecompressor.base import LlamaIndexCompressor
from autorag.utils.util import get_event_loop, process_batch


class TreeSummarize(LlamaIndexCompressor):
	def _pure(
		self,
		queries: List[str],
		contents: List[List[str]],
		prompt: Optional[str] = None,
		chat_prompt: Optional[str] = None,
		batch: int = 16,
	) -> List[str]:
		"""
		Recursively merge retrieved texts and summarizes them in a bottom-up fashion.
		This function is a wrapper for llama_index.response_synthesizers.TreeSummarize.
		For more information, visit https://docs.llamaindex.ai/en/latest/examples/response_synthesizers/tree_summarize.html.

		:param queries: The queries for retrieved passages.
		:param contents: The contents of retrieved passages.
		:param prompt: The prompt template for summarization.
		    If you want to use chat prompt, you should pass chat_prompt instead.
		    At prompt, you must specify where to put 'context_str' and 'query_str'.
		    Default is None. When it is None, it will use llama index default prompt.
		:param chat_prompt: The chat prompt template for summarization.
		    If you want to use normal prompt, you should pass prompt instead.
		    At prompt, you must specify where to put 'context_str' and 'query_str'.
		    Default is None. When it is None, it will use llama index default chat prompt.
		:param batch: The batch size for llm.
		    Set low if you face some errors.
		    Default is 16.
		:return: The list of compressed texts.
		"""
		if prompt is not None and not is_chat_model(self.llm):
			summary_template = PromptTemplate(prompt, prompt_type=PromptType.SUMMARY)
		elif chat_prompt is not None and is_chat_model(self.llm):
			summary_template = PromptTemplate(
				chat_prompt, prompt_type=PromptType.SUMMARY
			)
		else:
			summary_template = None
		summarizer = ts(llm=self.llm, summary_template=summary_template, use_async=True)
		tasks = [
			summarizer.aget_response(query, content)
			for query, content in zip(queries, contents)
		]
		loop = get_event_loop()
		results = loop.run_until_complete(process_batch(tasks, batch_size=batch))
		return results



import itertools
from typing import Dict, List

from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.base.llms.types import ChatResponse, ChatMessage, MessageRole

from autorag.data.qa.query.prompt import QUERY_GEN_PROMPT, QUERY_GEN_PROMPT_EXTRA


async def llama_index_generate_base(
	row: Dict,
	llm: BaseLLM,
	messages: List[ChatMessage],
) -> Dict:
	context = list(itertools.chain.from_iterable(row["retrieval_gt_contents"]))
	context_str = "\n".join([f"{i + 1}. {c}" for i, c in enumerate(context)])
	user_prompt = f"Text:\n{context_str}\n\nGenerated Question from the Text:\n"
	user_message = ChatMessage(role=MessageRole.USER, content=user_prompt)
	new_messages = [*messages, user_message]
	chat_response: ChatResponse = await llm.achat(messages=new_messages)
	row["query"] = chat_response.message.content
	return row


async def factoid_query_gen(
	row: Dict,
	llm: BaseLLM,
	lang: str = "en",
) -> Dict:
	return await llama_index_generate_base(
		row, llm, QUERY_GEN_PROMPT["factoid_single_hop"][lang]
	)


async def concept_completion_query_gen(
	row: Dict,
	llm: BaseLLM,
	lang: str = "en",
) -> Dict:
	return await llama_index_generate_base(
		row, llm, QUERY_GEN_PROMPT["concept_completion"][lang]
	)


async def two_hop_incremental(
	row: Dict,
	llm: BaseLLM,
	lang: str = "en",
) -> Dict:
	messages = QUERY_GEN_PROMPT["two_hop_incremental"][lang]
	passages = row["retrieval_gt_contents"]
	assert (
		len(passages) >= 2
	), "You have to sample more than two passages for making two-hop questions."
	context_str = f"Document 1: {passages[0][0]}\nDocument 2: {passages[1][0]}"
	user_prompt = f"{context_str}\n\nGenerated two-hop Question from two Documents:\n"
	messages.append(ChatMessage(role=MessageRole.USER, content=user_prompt))

	chat_response: ChatResponse = await llm.achat(messages=messages)
	response = chat_response.message.content
	row["query"] = response.split(":")[-1].strip()
	return row


async def custom_query_gen(
	row: Dict,
	llm: BaseLLM,
	messages: List[ChatMessage],
) -> Dict:
	return await llama_index_generate_base(row, llm, messages)


# Experimental feature: can only use factoid_single_hop
async def multiple_queries_gen(
	row: Dict,
	llm: BaseLLM,
	lang: str = "en",
	n: int = 3,
) -> Dict:
	_messages = QUERY_GEN_PROMPT["factoid_single_hop"][lang]
	_messages[0].content += QUERY_GEN_PROMPT_EXTRA["multiple_queries"][lang].format(n=n)
	return await llama_index_generate_base(row, llm, _messages)



import os
import pathlib
from typing import Tuple, List, Union, Dict

import pandas as pd

from autorag.evaluation import evaluate_retrieval
from autorag.schema.metricinput import MetricInput
from autorag.strategy import measure_speed, filter_by_threshold, select_best


def evaluate_retrieval_node(
	result_df: pd.DataFrame,
	metric_inputs: List[MetricInput],
	metrics: Union[List[str], List[Dict]],
) -> pd.DataFrame:
	"""
	Evaluate retrieval node from retrieval node result dataframe.
	:param result_df: The result dataframe from a retrieval node.
	:param metric_inputs: List of metric input schema for AutoRAG.
	:param metrics: Metric list from input strategies.
	:return: Return result_df with metrics columns.
	    The columns will be 'retrieved_contents', 'retrieved_ids', 'retrieve_scores', and metric names.
	"""

	@evaluate_retrieval(
		metric_inputs=metric_inputs,
		metrics=metrics,
	)
	def evaluate_this_module(df: pd.DataFrame):
		return (
			df["retrieved_contents"].tolist(),
			df["retrieved_ids"].tolist(),
			df["retrieve_scores"].tolist(),
		)

	return evaluate_this_module(result_df)


def run(
	input_modules,
	input_module_params,
	project_dir: Union[str, pathlib.Path, pathlib.PurePath],
	previous_result: pd.DataFrame,
	strategies,
	metric_inputs: List[MetricInput],
) -> Tuple[List[pd.DataFrame], List]:
	"""
	Run input modules and parameters.
	:param input_modules: Input modules
	:param input_module_params: Input module parameters
	:param project_dir: Project directory path.
	:param previous_result: Previous result dataframe.
	:param strategies: Strategies for retrieval node.
	:param metric_inputs: List of metric input schema for AutoRAG.
	:return: First, it returns list of result dataframe.
	Second, it returns list of execution times.
	"""
	result, execution_times = zip(
		*map(
			lambda task: measure_speed(
				task[0].run_evaluator,
				project_dir=project_dir,
				previous_result=previous_result,
				**task[1],
			),
			zip(input_modules, input_module_params),
		)
	)
	average_times = list(map(lambda x: x / len(result[0]), execution_times))

	# run metrics before filtering
	if strategies.get("metrics") is None:
		raise ValueError("You must at least one metrics for retrieval evaluation.")
	result = list(
		map(
			lambda x: evaluate_retrieval_node(
				x,
				metric_inputs,
				strategies.get("metrics"),
			),
			result,
		)
	)

	return result, average_times


def save_and_summary(
	input_modules,
	input_module_params,
	result_list,
	execution_time_list,
	filename_start: int,
	save_dir: Union[str, pathlib.Path, pathlib.PurePath],
	strategies,
):
	"""
	Save the result and make summary file
	:param input_modules: Input modules
	:param input_module_params: Input module parameters
	:param result_list: Result list
	:param execution_time_list: Execution times
	:param filename_start: The first filename to use
	:return: First, it returns list of result dataframe.
	Second, it returns list of execution times.
	"""

	# save results to folder
	filepaths = list(
		map(
			lambda x: os.path.join(save_dir, f"{x}.parquet"),
			range(filename_start, filename_start + len(input_modules)),
		)
	)
	list(
		map(
			lambda x: x[0].to_parquet(x[1], index=False),
			zip(result_list, filepaths),
		)
	)  # execute save to parquet
	filename_list = list(map(lambda x: os.path.basename(x), filepaths))

	summary_df = pd.DataFrame(
		{
			"filename": filename_list,
			"module_name": list(map(lambda module: module.__name__, input_modules)),
			"module_params": input_module_params,
			"execution_time": execution_time_list,
			**{
				metric: list(map(lambda result: result[metric].mean(), result_list))
				for metric in strategies.get("metrics")
			},
		}
	)
	summary_df.to_csv(os.path.join(save_dir, "summary.csv"), index=False)
	return summary_df


def find_best(results, average_times, filenames, strategies):
	# filter by strategies
	if strategies.get("speed_threshold") is not None:
		results, filenames = filter_by_threshold(
			results, average_times, strategies["speed_threshold"], filenames
		)
	selected_result, selected_filename = select_best(
		results,
		strategies.get("metrics"),
		filenames,
		strategies.get("strategy", "mean"),
	)
	return selected_result, selected_filename



import logging

from qdrant_client import QdrantClient
from qdrant_client.models import (
	Distance,
	VectorParams,
	PointStruct,
	PointIdsList,
	HasIdCondition,
	Filter,
	SearchRequest,
)

from typing import List, Tuple, Union

from autorag.vectordb import BaseVectorStore

logger = logging.getLogger("AutoRAG")


class Qdrant(BaseVectorStore):
	def __init__(
		self,
		embedding_model: Union[str, List[dict]],
		collection_name: str,
		embedding_batch: int = 100,
		similarity_metric: str = "cosine",
		client_type: str = "docker",
		url: str = "http://localhost:6333",
		host: str = "",
		api_key: str = "",
		dimension: int = 1536,
		ingest_batch: int = 64,
		parallel: int = 1,
		max_retries: int = 3,
	):
		super().__init__(embedding_model, similarity_metric, embedding_batch)

		self.collection_name = collection_name
		self.ingest_batch = ingest_batch
		self.parallel = parallel
		self.max_retries = max_retries

		if similarity_metric == "cosine":
			distance = Distance.COSINE
		elif similarity_metric == "ip":
			distance = Distance.DOT
		elif similarity_metric == "l2":
			distance = Distance.EUCLID
		else:
			raise ValueError(
				f"similarity_metric {similarity_metric} is not supported\n"
				"supported similarity metrics are: cosine, ip, l2"
			)

		if client_type == "docker":
			self.client = QdrantClient(
				url=url,
			)
		elif client_type == "cloud":
			self.client = QdrantClient(
				host=host,
				api_key=api_key,
			)
		else:
			raise ValueError(
				f"client_type {client_type} is not supported\n"
				"supported client types are: docker, cloud"
			)

		if not self.client.collection_exists(collection_name):
			self.client.create_collection(
				collection_name,
				vectors_config=VectorParams(
					size=dimension,
					distance=distance,
				),
			)
		self.collection = self.client.get_collection(collection_name)

	async def add(self, ids: List[str], texts: List[str]):
		texts = self.truncated_inputs(texts)
		text_embeddings = await self.embedding.aget_text_embedding_batch(texts)
		self.add_embedding(ids, text_embeddings)

	def add_embedding(self, ids: List[str], embeddings: List[List[float]]):
		points = list(
			map(lambda x: PointStruct(id=x[0], vector=x[1]), zip(ids, embeddings))
		)

		self.client.upload_points(
			collection_name=self.collection_name,
			points=points,
			batch_size=self.ingest_batch,
			parallel=self.parallel,
			max_retries=self.max_retries,
			wait=True,
		)

	async def fetch(self, ids: List[str]) -> List[List[float]]:
		# Fetch vectors by IDs
		fetched_results = self.client.retrieve(
			collection_name=self.collection_name,
			ids=ids,
			with_vectors=True,
		)
		return list(map(lambda x: x.vector, fetched_results))

	async def is_exist(self, ids: List[str]) -> List[bool]:
		existed_result = self.client.scroll(
			collection_name=self.collection_name,
			scroll_filter=Filter(
				must=[
					HasIdCondition(has_id=ids),
				],
			),
		)
		# existed_result is tuple. So we use existed_result[0] to get list of Record
		existed_ids = list(map(lambda x: x.id, existed_result[0]))
		return list(map(lambda x: x in existed_ids, ids))

	async def query(
		self, queries: List[str], top_k: int, **kwargs
	) -> Tuple[List[List[str]], List[List[float]]]:
		queries = self.truncated_inputs(queries)
		query_embeddings: List[
			List[float]
		] = await self.embedding.aget_text_embedding_batch(queries)

		search_queries = list(
			map(
				lambda x: SearchRequest(vector=x, limit=top_k, with_vector=True),
				query_embeddings,
			)
		)

		search_result = self.client.search_batch(
			collection_name=self.collection_name, requests=search_queries
		)

		# Extract IDs and distances
		ids = [[str(hit.id) for hit in result] for result in search_result]
		scores = [[hit.score for hit in result] for result in search_result]

		return ids, scores

	async def delete(self, ids: List[str]):
		self.client.delete(
			collection_name=self.collection_name,
			points_selector=PointIdsList(points=ids),
		)

	def delete_collection(self):
		# Delete the collection
		self.client.delete_collection(self.collection_name)



import os
from typing import List

from autorag.support import dynamically_find_function
from autorag.utils.util import load_yaml_config
from autorag.vectordb.base import BaseVectorStore


def get_support_vectordb(vectordb_name: str):
	support_vectordb = {
		"chroma": ("autorag.vectordb.chroma", "Chroma"),
		"Chroma": ("autorag.vectordb.chroma", "Chroma"),
		"milvus": ("autorag.vectordb.milvus", "Milvus"),
		"Milvus": ("autorag.vectordb.milvus", "Milvus"),
		"weaviate": ("autorag.vectordb.weaviate", "Weaviate"),
		"Weaviate": ("autorag.vectordb.weaviate", "Weaviate"),
		"pinecone": ("autorag.vectordb.pinecone", "Pinecone"),
		"Pinecone": ("autorag.vectordb.pinecone", "Pinecone"),
		"couchbase": ("autorag.vectordb.couchbase", "Couchbase"),
		"Couchbase": ("autorag.vectordb.couchbase", "Couchbase"),
		"qdrant": ("autorag.vectordb.qdrant", "Qdrant"),
		"Qdrant": ("autorag.vectordb.qdrant", "Qdrant"),
	}
	return dynamically_find_function(vectordb_name, support_vectordb)


def load_vectordb(vectordb_name: str, **kwargs):
	vectordb = get_support_vectordb(vectordb_name)
	return vectordb(**kwargs)


def load_vectordb_from_yaml(yaml_path: str, vectordb_name: str, project_dir: str):
	config_dict = load_yaml_config(yaml_path)
	vectordb_list = config_dict.get("vectordb", [])
	if len(vectordb_list) == 0 or vectordb_name == "default":
		chroma_path = os.path.join(project_dir, "resources", "chroma")
		return load_vectordb(
			"chroma",
			client_type="persistent",
			embedding_model="openai",
			collection_name="openai",
			path=chroma_path,
		)

	target_dict = list(filter(lambda x: x["name"] == vectordb_name, vectordb_list))
	target_dict[0].pop("name")  # delete a name key
	target_vectordb_name = target_dict[0].pop("db_type")
	target_vectordb_params = target_dict[0]
	return load_vectordb(target_vectordb_name, **target_vectordb_params)


def load_all_vectordb_from_yaml(
	yaml_path: str, project_dir: str
) -> List[BaseVectorStore]:
	config_dict = load_yaml_config(yaml_path)
	vectordb_list = config_dict.get("vectordb", [])
	if len(vectordb_list) == 0:
		chroma_path = os.path.join(project_dir, "resources", "chroma")
		return [
			load_vectordb(
				"chroma",
				client_type="persistent",
				embedding_model="openai",
				collection_name="openai",
				path=chroma_path,
			)
		]

	result_vectordbs = []
	for vectordb_dict in vectordb_list:
		_ = vectordb_dict.pop("name")
		vectordb_type = vectordb_dict.pop("db_type")
		vectordb = load_vectordb(vectordb_type, **vectordb_dict)
		result_vectordbs.append(vectordb)
	return result_vectordbs



import itertools
from typing import Dict, List

from llama_index.core.base.llms.types import ChatMessage, MessageRole
from llama_index.llms.openai.utils import to_openai_message_dicts
from openai import AsyncClient
from pydantic import BaseModel

from autorag.data.qa.evolve.prompt import QUERY_EVOLVE_PROMPT


class Response(BaseModel):
	evolved_query: str


async def query_evolve_openai_base(
	row: Dict,
	client: AsyncClient,
	messages: List[ChatMessage],
	model_name: str = "gpt-4o-2024-08-06",
):
	"""
	Evolve the original query to a new evolved query using OpenAI structured outputs.
	"""
	original_query = row["query"]
	context = list(itertools.chain.from_iterable(row["retrieval_gt_contents"]))
	context_str = "Text:\n" + "\n".join(
		[f"{i + 1}. {c}" for i, c in enumerate(context)]
	)
	user_prompt = f"Question: {original_query}\nContext: {context_str}\nOutput: "
	messages.append(ChatMessage(role=MessageRole.USER, content=user_prompt))

	completion = await client.beta.chat.completions.parse(
		model=model_name,
		messages=to_openai_message_dicts(messages),
		response_format=Response,
	)
	row["query"] = completion.choices[0].message.parsed.evolved_query
	return row


async def conditional_evolve_ragas(
	row: Dict,
	client: AsyncClient,
	model_name: str = "gpt-4o-2024-08-06",
	lang: str = "en",
) -> Dict:
	return await query_evolve_openai_base(
		row, client, QUERY_EVOLVE_PROMPT["conditional_evolve_ragas"][lang], model_name
	)


async def reasoning_evolve_ragas(
	row: Dict,
	client: AsyncClient,
	model_name: str = "gpt-4o-2024-08-06",
	lang: str = "en",
) -> Dict:
	return await query_evolve_openai_base(
		row, client, QUERY_EVOLVE_PROMPT["reasoning_evolve_ragas"][lang], model_name
	)


async def compress_ragas(
	row: Dict,
	client: AsyncClient,
	model_name: str = "gpt-4o-2024-08-06",
	lang: str = "en",
) -> Dict:
	original_query = row["query"]
	messages = QUERY_EVOLVE_PROMPT["compress_ragas"][lang]
	user_prompt = f"Question: {original_query}\nOutput: "
	messages.append(ChatMessage(role=MessageRole.USER, content=user_prompt))

	completion = await client.beta.chat.completions.parse(
		model=model_name,
		messages=to_openai_message_dicts(messages),
		response_format=Response,
	)
	row["query"] = completion.choices[0].message.parsed.evolved_query
	return row



import base64
import itertools
import json
import os
from typing import List, Optional, Tuple

import aiohttp
import fitz  # PyMuPDF

from autorag.data.parse.base import parser_node
from autorag.utils.util import process_batch, get_event_loop


@parser_node
def clova_ocr(
	data_path_list: List[str],
	url: Optional[str] = None,
	api_key: Optional[str] = None,
	batch: int = 5,
	table_detection: bool = False,
) -> Tuple[List[str], List[str], List[int]]:
	"""
	Parse documents to use Naver Clova OCR.

	:param data_path_list: The list of data paths to parse.
	:param url: The URL for Clova OCR.
		You can get the URL with the guide at https://guide.ncloud-docs.com/docs/clovaocr-example01
		You can set the environment variable CLOVA_URL, or you can set it directly as a parameter.
	:param api_key: The API key for Clova OCR.
		You can get the API key with the guide at https://guide.ncloud-docs.com/docs/clovaocr-example01
		You can set the environment variable CLOVA_API_KEY, or you can set it directly as a parameter.
	:param batch: The batch size for parse documents. Default is 8.
	:param table_detection: Whether to enable table detection. Default is False.
	:return: tuple of lists containing the parsed texts, path and pages.
	"""
	url = os.getenv("CLOVA_URL", None) if url is None else url
	if url is None:
		raise KeyError(
			"Please set the URL for Clova OCR in the environment variable CLOVA_URL "
			"or directly set it on the config YAML file."
		)

	api_key = os.getenv("CLOVA_API_KEY", None) if api_key is None else api_key
	if api_key is None:
		raise KeyError(
			"Please set the API key for Clova OCR in the environment variable CLOVA_API_KEY "
			"or directly set it on the config YAML file."
		)
	if batch > 5:
		raise ValueError("The batch size should be less than or equal to 5.")

	image_data_lst = list(
		map(lambda data_path: pdf_to_images(data_path), data_path_list)
	)
	image_info_lst = [
		generate_image_info(pdf_path, len(image_data))
		for pdf_path, image_data in zip(data_path_list, image_data_lst)
	]

	image_data_list = list(itertools.chain(*image_data_lst))
	image_info_list = list(itertools.chain(*image_info_lst))

	tasks = [
		clova_ocr_pure(image_data, image_info, url, api_key, table_detection)
		for image_data, image_info in zip(image_data_list, image_info_list)
	]
	loop = get_event_loop()
	results = loop.run_until_complete(process_batch(tasks, batch))

	texts, path, pages = zip(*results)
	return list(texts), list(path), list(pages)


async def clova_ocr_pure(
	image_data: bytes,
	image_info: dict,
	url: str,
	api_key: str,
	table_detection: bool = False,
) -> Tuple[str, str, int]:
	session = aiohttp.ClientSession()
	table_html = ""
	headers = {"X-OCR-SECRET": api_key, "Content-Type": "application/json"}

	# Convert image data to base64
	image_base64 = base64.b64encode(image_data).decode("utf-8")

	# Set data
	data = {
		"version": "V2",
		"requestId": "sample_id",
		"timestamp": 0,
		"images": [{"format": "png", "name": "sample_image", "data": image_base64}],
		"enableTableDetection": table_detection,
	}

	async with session.post(url, headers=headers, data=json.dumps(data)) as response:
		resp_json = await response.json()
		if "images" not in resp_json:
			raise RuntimeError(
				f"Invalid response from Clova API: {resp_json['detail']}"
			)
		if "tables" in resp_json["images"][0].keys():
			table_html = json_to_html_table(
				resp_json["images"][0]["tables"][0]["cells"]
			)
		page_text = extract_text_from_fields(resp_json["images"][0]["fields"])

		if table_html:
			page_text += f"\n\ntable html:\n{table_html}"

		await session.close()
		return page_text, image_info["pdf_path"], image_info["pdf_page"]


def pdf_to_images(pdf_path: str) -> List[bytes]:
	"""Convert each page of the PDF to an image and return the image data."""
	pdf_document = fitz.open(pdf_path)
	image_data_lst = []
	for page_num in range(len(pdf_document)):
		page = pdf_document.load_page(page_num)
		pix = page.get_pixmap()
		img_data = pix.tobytes("png")
		image_data_lst.append(img_data)
	return image_data_lst


def generate_image_info(pdf_path: str, num_pages: int) -> List[dict]:
	"""Generate image names based on the PDF file name and the number of pages."""
	image_info_lst = [
		{"pdf_path": pdf_path, "pdf_page": page_num + 1}
		for page_num in range(num_pages)
	]
	return image_info_lst


def extract_text_from_fields(fields):
	text = ""
	for field in fields:
		text += field["inferText"]
		if field["lineBreak"]:
			text += "\n"
		else:
			text += " "
	return text.strip()


def json_to_html_table(json_data):
	# Initialize the HTML table
	html = '<table border="1">\n'
	# Determine the number of rows and columns
	max_row = max(cell["rowIndex"] + cell["rowSpan"] for cell in json_data)
	max_col = max(cell["columnIndex"] + cell["columnSpan"] for cell in json_data)
	# Create a 2D array to keep track of merged cells
	table = [["" for _ in range(max_col)] for _ in range(max_row)]
	# Fill the table with cell data
	for cell in json_data:
		row = cell["rowIndex"]
		col = cell["columnIndex"]
		row_span = cell["rowSpan"]
		col_span = cell["columnSpan"]
		cell_text = (
			" ".join(
				line["inferText"] for line in cell["cellTextLines"][0]["cellWords"]
			)
			if cell["cellTextLines"]
			else ""
		)
		# Place the cell in the table
		table[row][col] = {"text": cell_text, "rowSpan": row_span, "colSpan": col_span}
		# Mark merged cells as occupied
		for r in range(row, row + row_span):
			for c in range(col, col + col_span):
				if r != row or c != col:
					table[r][c] = None
	# Generate HTML from the table array
	for row in table:
		html += "  <tr>\n"
		for cell in row:
			if cell is None:
				continue
			if cell == "":
				html += "    <td></td>\n"
			else:
				row_span_attr = (
					f' rowspan="{cell["rowSpan"]}"' if cell["rowSpan"] > 1 else ""
				)
				col_span_attr = (
					f' colspan="{cell["colSpan"]}"' if cell["colSpan"] > 1 else ""
				)
				html += f'    <td{row_span_attr}{col_span_attr}>{cell["text"]}</td>\n'
		html += "  </tr>\n"
	html += "</table>"
	return html



from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pandas as pd

from autorag.nodes.hybridretrieval.base import HybridRetrieval
from autorag.nodes.hybridretrieval.run import evaluate_retrieval_node
from autorag.strategy import select_best


class HybridRRF(HybridRetrieval):
	def _pure(self, info: dict, top_k: int, weight: int = 60, rrf_k: int = -1):
		return hybrid_rrf(
			(info["retrieved_ids_semantic"], info["retrieved_ids_lexical"]),
			(info["retrieve_scores_semantic"], info["retrieve_scores_lexical"]),
			top_k,
			weight,
			rrf_k,
		)

	@classmethod
	def run_evaluator(
		cls,
		project_dir: Union[str, Path],
		previous_result: pd.DataFrame,
		*args,
		**kwargs,
	):
		assert "strategy" in kwargs, "You must specify the strategy to use."
		assert (
			"input_metrics" in kwargs
		), "You must specify the input metrics to use, which is list of MetricInput."
		strategies = kwargs.pop("strategy")
		input_metrics = kwargs.pop("input_metrics")
		weight_range = kwargs.pop("weight_range", (4, 80))
		test_weight_size = weight_range[1] - weight_range[0] + 1
		weight_candidates = np.linspace(
			weight_range[0], weight_range[1], test_weight_size
		).tolist()

		result_list = []
		instance = cls(project_dir, *args, **kwargs)
		for weight_value in weight_candidates:
			result_df = instance.pure(previous_result, weight=weight_value, **kwargs)
			result_list.append(result_df)

		if strategies.get("metrics") is None:
			raise ValueError("You must at least one metrics for retrieval evaluation.")
		result_list = list(
			map(
				lambda x: evaluate_retrieval_node(
					x,
					input_metrics,
					strategies.get("metrics"),
				),
				result_list,
			)
		)

		# select best result
		best_result_df, best_weight = select_best(
			result_list,
			strategies.get("metrics"),
			metadatas=weight_candidates,
			strategy_name=strategies.get("strategy", "normalize_mean"),
		)
		return {
			"best_result": best_result_df,
			"best_weight": best_weight,
		}


def hybrid_rrf(
	ids: Tuple,
	scores: Tuple,
	top_k: int,
	weight: int = 60,
	rrf_k: int = -1,
) -> Tuple[List[List[str]], List[List[float]]]:
	"""
	Hybrid RRF function.
	RRF (Rank Reciprocal Fusion) is a method to fuse multiple retrieval results.
	It is common to fuse dense retrieval and sparse retrieval results using RRF.
	To use this function, you must input ids and scores as tuple.
	It is more unique than other retrieval modules because it does not really execute retrieval but just fuses
	the results of other retrieval functions.
	So you have to run more than two retrieval modules before running this function.
	And collect ids and scores result from each retrieval module.
	Make it as a tuple and input it to this function.

	:param ids: The tuple of ids that you want to fuse.
	    The length of this must be the same as the length of scores.
	:param scores: The retrieve scores that you want to fuse.
	    The length of this must be the same as the length of ids.
	:param top_k: The number of passages to be retrieved.
	:param weight: Hyperparameter for RRF.
	    It was originally rrf_k value.
	    Default is 60.
	    For more information, please visit our documentation.
	:param rrf_k: (Deprecated) Hyperparameter for RRF.
	    It was originally rrf_k value. Will remove at a further version.
	:return: The tuple of ids and fused scores that are fused by RRF.
	"""
	assert len(ids) == len(scores), "The length of ids and scores must be the same."
	assert len(ids) > 1, "You must input more than one retrieval results."
	assert top_k > 0, "top_k must be greater than 0."
	assert weight > 0, "rrf_k must be greater than 0."

	if rrf_k != -1:
		weight = int(rrf_k)
	else:
		weight = int(weight)

	id_df = pd.DataFrame({f"id_{i}": id_list for i, id_list in enumerate(ids)})
	score_df = pd.DataFrame(
		{f"score_{i}": score_list for i, score_list in enumerate(scores)}
	)
	df = pd.concat([id_df, score_df], axis=1)

	def rrf_pure_apply(row):
		ids_tuple = tuple(row[[f"id_{i}" for i in range(len(ids))]].values)
		scores_tuple = tuple(row[[f"score_{i}" for i in range(len(scores))]].values)
		return pd.Series(rrf_pure(ids_tuple, scores_tuple, weight, top_k))

	df[["rrf_id", "rrf_score"]] = df.apply(rrf_pure_apply, axis=1)
	return df["rrf_id"].tolist(), df["rrf_score"].tolist()


def rrf_pure(
	ids: Tuple, scores: Tuple, rrf_k: int, top_k: int
) -> Tuple[List[str], List[float]]:
	df = pd.concat(
		[pd.Series(dict(zip(_id, score))) for _id, score in zip(ids, scores)], axis=1
	)
	rank_df = df.rank(ascending=False, method="min")
	rank_df = rank_df.fillna(0)
	rank_df["rrf"] = rank_df.apply(lambda row: rrf_calculate(row, rrf_k), axis=1)
	rank_df = rank_df.sort_values(by="rrf", ascending=False)
	return rank_df.index.tolist()[:top_k], rank_df["rrf"].tolist()[:top_k]


def rrf_calculate(row, rrf_k):
	result = 0
	for r in row:
		if r == 0:
			continue
		result += 1 / (r + rrf_k)
	return result



import logging
import os
import pathlib
import uuid
from typing import Dict, Optional, List, Union, Literal

import pandas as pd
from quart import Quart, request, jsonify
from quart.helpers import stream_with_context
from pydantic import BaseModel, ValidationError

from autorag.deploy.base import BaseRunner
from autorag.nodes.generator.base import BaseGenerator
from autorag.nodes.promptmaker.base import BasePromptMaker
from autorag.utils.util import fetch_contents, to_list

logger = logging.getLogger("AutoRAG")

deploy_dir = pathlib.Path(__file__).parent
root_dir = pathlib.Path(__file__).parent.parent

VERSION_PATH = os.path.join(root_dir, "VERSION")


class QueryRequest(BaseModel):
	query: str
	result_column: Optional[str] = "generated_texts"


class RetrievedPassage(BaseModel):
	content: str
	doc_id: str
	score: float
	filepath: Optional[str] = None
	file_page: Optional[int] = None
	start_idx: Optional[int] = None
	end_idx: Optional[int] = None


class RunResponse(BaseModel):
	result: Union[str, List[str]]
	retrieved_passage: List[RetrievedPassage]


class RetrievalResponse(BaseModel):
	passages: List[RetrievedPassage]


class StreamResponse(BaseModel):
	"""
	When the type is generated_text, only generated_text is returned. The other fields are None.
	When the type is retrieved_passage, only retrieved_passage and passage_index are returned. The other fields are None.
	"""

	type: Literal["generated_text", "retrieved_passage"]
	generated_text: Optional[str]
	retrieved_passage: Optional[RetrievedPassage]
	passage_index: Optional[int]


class VersionResponse(BaseModel):
	version: str


class ApiRunner(BaseRunner):
	def __init__(self, config: Dict, project_dir: Optional[str] = None):
		super().__init__(config, project_dir)
		self.app = Quart(__name__)

		data_dir = os.path.join(project_dir, "data")
		self.corpus_df = pd.read_parquet(
			os.path.join(data_dir, "corpus.parquet"), engine="pyarrow"
		)
		self.__add_api_route()

	def __add_api_route(self):
		@self.app.route("/v1/run", methods=["POST"])
		async def run_query():
			try:
				data = await request.get_json()
				data = QueryRequest(**data)
			except ValidationError as e:
				return jsonify(e.errors()), 400

			previous_result = pd.DataFrame(
				{
					"qid": str(uuid.uuid4()),
					"query": [data.query],
					"retrieval_gt": [[]],
					"generation_gt": [""],
				}
			)  # pseudo qa data for execution
			for module_instance, module_param in zip(
				self.module_instances, self.module_params
			):
				new_result = module_instance.pure(
					previous_result=previous_result, **module_param
				)
				duplicated_columns = previous_result.columns.intersection(
					new_result.columns
				)
				drop_previous_result = previous_result.drop(columns=duplicated_columns)
				previous_result = pd.concat([drop_previous_result, new_result], axis=1)

			# Simulate processing the query
			generated_text = previous_result[data.result_column].tolist()[0]
			retrieved_passage = self.extract_retrieve_passage(previous_result)

			response = RunResponse(
				result=generated_text, retrieved_passage=retrieved_passage
			)

			return jsonify(response.model_dump()), 200

		@self.app.route("/v1/retrieve", methods=["POST"])
		async def run_retrieve_only():
			data = await request.get_json()
			query = data.get("query", None)
			if query is None:
				return jsonify(
					{
						"error": "Invalid request. You need to include 'query' in the request body."
					}
				), 400

			previous_result = pd.DataFrame(
				{
					"qid": str(uuid.uuid4()),
					"query": [query],
					"retrieval_gt": [[]],
					"generation_gt": [""],
				}
			)  # pseudo qa data for execution
			for module_instance, module_param in zip(
				self.module_instances, self.module_params
			):
				if isinstance(module_instance, BasePromptMaker) or isinstance(
					module_instance, BaseGenerator
				):
					continue
				new_result = module_instance.pure(
					previous_result=previous_result, **module_param
				)
				duplicated_columns = previous_result.columns.intersection(
					new_result.columns
				)
				drop_previous_result = previous_result.drop(columns=duplicated_columns)
				previous_result = pd.concat([drop_previous_result, new_result], axis=1)

			# Simulate processing the query
			retrieved_passages = self.extract_retrieve_passage(previous_result)

			retrieval_response = RetrievalResponse(passages=retrieved_passages)
			return jsonify(retrieval_response.model_dump()), 200

		@self.app.route("/v1/stream", methods=["POST"])
		async def stream_query():
			try:
				data = await request.get_json()
				data = QueryRequest(**data)
			except ValidationError as e:
				return jsonify(e.errors()), 400

			@stream_with_context
			async def generate():
				previous_result = pd.DataFrame(
					{
						"qid": str(uuid.uuid4()),
						"query": [data.query],
						"retrieval_gt": [[]],
						"generation_gt": [""],
					}
				)  # pseudo qa data for execution

				for module_instance, module_param in zip(
					self.module_instances, self.module_params
				):
					if not isinstance(module_instance, BaseGenerator):
						new_result = module_instance.pure(
							previous_result=previous_result, **module_param
						)
						duplicated_columns = previous_result.columns.intersection(
							new_result.columns
						)
						drop_previous_result = previous_result.drop(
							columns=duplicated_columns
						)
						previous_result = pd.concat(
							[drop_previous_result, new_result], axis=1
						)
					else:
						retrieved_passages = self.extract_retrieve_passage(
							previous_result
						)
						for i, retrieved_passage in enumerate(retrieved_passages):
							yield (
								StreamResponse(
									type="retrieved_passage",
									generated_text=None,
									retrieved_passage=retrieved_passage,
									passage_index=i,
								)
								.model_dump_json()
								.encode("utf-8")
							)
						# Start streaming of the result
						assert len(previous_result) == 1
						prompt: str = previous_result["prompts"].tolist()[0]
						async for delta in module_instance.astream(
							prompt=prompt, **module_param
						):
							response = StreamResponse(
								type="generated_text",
								generated_text=delta,
								retrieved_passage=None,
								passage_index=None,
							)
							yield response.model_dump_json().encode("utf-8")

			return generate(), 200, {"X-Something": "value"}

		@self.app.route("/version", methods=["GET"])
		def get_version():
			with open(VERSION_PATH, "r") as f:
				version = f.read().strip()
			response = VersionResponse(version=version)
			return jsonify(response.model_dump()), 200

	def run_api_server(
		self, host: str = "0.0.0.0", port: int = 8000, remote: bool = True, **kwargs
	):
		"""
		Run the pipeline as an api server.
		Here is api endpoint documentation => https://marker-inc-korea.github.io/AutoRAG/deploy/api_endpoint.html

		:param host: The host of the api server.
		:param port: The port of the api server.
		:param remote: Whether to expose the api server to the public internet using ngrok.
		:param kwargs: Other arguments for Flask app.run.
		"""
		logger.info(f"Run api server at {host}:{port}")
		if remote:
			from pyngrok import ngrok

			http_tunnel = ngrok.connect(str(port), "http")
			public_url = http_tunnel.public_url
			logger.info(f"Public API URL: {public_url}")
		self.app.run(host=host, port=port, **kwargs)

	def extract_retrieve_passage(self, df: pd.DataFrame) -> List[RetrievedPassage]:
		if "retrieved_ids" not in df.columns and "retrieved_ids_semantic" in df.columns:
			retrieved_ids: List[str] = df["retrieved_ids_semantic"].tolist()[0]
			scores = df["retrieve_scores_semantic"].tolist()[0]
		elif (
			"retrieved_ids" not in df.columns
			and "retrieved_ids_semantic" not in df.columns
		):
			retrieved_ids: List[str] = df["retrieved_ids_lexical"].tolist()[0]
			scores = df["retrieve_scores_lexical"].tolist()[0]
		else:
			retrieved_ids: List[str] = df["retrieved_ids"].tolist()[0]
			scores = df["retrieve_scores"].tolist()[0]
		contents = fetch_contents(self.corpus_df, [retrieved_ids])[0]
		if "path" in self.corpus_df.columns:
			paths = fetch_contents(self.corpus_df, [retrieved_ids], column_name="path")[
				0
			]
		else:
			paths = [None] * len(retrieved_ids)
		metadatas = fetch_contents(
			self.corpus_df, [retrieved_ids], column_name="metadata"
		)[0]
		if "start_end_idx" in self.corpus_df.columns:
			start_end_indices = fetch_contents(
				self.corpus_df, [retrieved_ids], column_name="start_end_idx"
			)[0]
		else:
			start_end_indices = [None] * len(retrieved_ids)
		start_end_indices = to_list(start_end_indices)
		return list(
			map(
				lambda content,
				doc_id,
				score,
				path,
				metadata,
				start_end_idx: RetrievedPassage(
					content=content,
					doc_id=doc_id,
					score=score,
					filepath=path,
					file_page=metadata.get("page", None),
					start_idx=start_end_idx[0] if start_end_idx else None,
					end_idx=start_end_idx[1] if start_end_idx else None,
				),
				contents,
				retrieved_ids,
				scores,
				paths,
				metadatas,
				start_end_indices,
			)
		)



"""
This file contains the retrieval contents metric,
which means calculate the metric based on the contents of the retrieved items.
"""

import itertools
from collections import Counter

import numpy as np

from autorag.evaluation.metric.util import autorag_metric
from autorag.schema.metricinput import MetricInput
from autorag.utils.util import normalize_string


def single_token_f1(ground_truth: str, prediction: str):
	prediction_tokens = normalize_string(prediction).split()
	ground_truth_tokens = normalize_string(ground_truth).split()
	common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
	num_same = sum(common.values())
	if num_same == 0:
		return 0, 0, 0
	precision = 1.0 * num_same / len(prediction_tokens)
	recall = 1.0 * num_same / len(ground_truth_tokens)
	f1 = (2 * precision * recall) / (precision + recall)
	return precision, recall, f1


@autorag_metric(fields_to_check=["retrieved_contents", "retrieval_gt_contents"])
def retrieval_token_f1(metric_input: MetricInput):
	pred = metric_input.retrieved_contents
	gt = itertools.chain.from_iterable(metric_input.retrieval_gt_contents)

	calculated_results = list(
		map(lambda x: single_token_f1(x[1], x[0]), list(itertools.product(pred, gt)))
	)
	_, _, result = zip(*calculated_results)
	result_np = np.array(list(result)).reshape(len(pred), -1)
	return result_np.max(axis=1).mean()


@autorag_metric(fields_to_check=["retrieved_contents", "retrieval_gt_contents"])
def retrieval_token_precision(metric_input: MetricInput):
	pred = metric_input.retrieved_contents
	gt = itertools.chain.from_iterable(metric_input.retrieval_gt_contents)

	calculated_results = list(
		map(lambda x: single_token_f1(x[1], x[0]), list(itertools.product(pred, gt)))
	)
	result, _, _ = zip(*calculated_results)
	result_np = np.array(list(result)).reshape(len(pred), -1)
	return result_np.max(axis=1).mean()


@autorag_metric(fields_to_check=["retrieved_contents", "retrieval_gt_contents"])
def retrieval_token_recall(metric_input: MetricInput):
	pred = metric_input.retrieved_contents
	gt = itertools.chain.from_iterable(metric_input.retrieval_gt_contents)

	calculated_results = list(
		map(lambda x: single_token_f1(x[1], x[0]), list(itertools.product(pred, gt)))
	)
	_, result, _ = zip(*calculated_results)
	result_np = np.array(list(result)).reshape(len(pred), -1)
	return result_np.max(axis=1).mean()



def cast_retrieve_infos(previous_result):
	return {
		"retrieved_contents": cast_retrieved_contents(previous_result),
		"retrieved_ids": cast_retrieved_ids(previous_result),
		"retrieve_scores": cast_retrieve_scores(previous_result),
	}


def cast_retrieved_contents(previous_result):
	if "retrieved_contents" in previous_result.columns:
		return previous_result["retrieved_contents"].tolist()
	elif "retrieved_contents_semantic" in previous_result.columns:
		return previous_result["retrieved_contents_semantic"].tolist()
	elif "retrieved_contents_lexical" in previous_result.columns:
		return previous_result["retrieved_contents_lexical"].tolist()
	else:
		raise ValueError(
			"previous_result must contain either 'retrieved_contents', 'retrieved_contents_semantic', or 'retrieved_contents_lexical' columns."
		)


def cast_retrieved_ids(previous_result):
	if "retrieved_ids" in previous_result.columns:
		return previous_result["retrieved_ids"].tolist()
	elif "retrieved_ids_semantic" in previous_result.columns:
		return previous_result["retrieved_ids_semantic"].tolist()
	elif "retrieved_ids_lexical" in previous_result.columns:
		return previous_result["retrieved_ids_lexical"].tolist()
	else:
		raise ValueError(
			"previous_result must contain either 'retrieved_ids', 'retrieved_ids_semantic', or 'retrieved_ids_lexical' columns."
		)


def cast_retrieve_scores(previous_result):
	if "retrieve_scores" in previous_result.columns:
		return previous_result["retrieve_scores"].tolist()
	elif "retrieve_scores_semantic" in previous_result.columns:
		return previous_result["retrieve_scores_semantic"].tolist()
	elif "retrieve_scores_lexical" in previous_result.columns:
		return previous_result["retrieve_scores_lexical"].tolist()
	else:
		raise ValueError(
			"previous_result must contain either 'retrieve_scores', 'retrieve_scores_semantic', or 'retrieve_scores_lexical' columns."
		)



import functools
import logging
from datetime import datetime
from glob import glob
from typing import Tuple, List, Optional
import os

from autorag.utils import result_to_dataframe
from autorag.data.utils.util import get_file_metadata

logger = logging.getLogger("AutoRAG")


def parser_node(func):
	@functools.wraps(func)
	@result_to_dataframe(["texts", "path", "page", "last_modified_datetime"])
	def wrapper(
		data_path_glob: str,
		file_type: str,
		parse_method: Optional[str] = None,
		**kwargs,
	) -> Tuple[List[str], List[str], List[int], List[datetime]]:
		logger.info(f"Running parser - {func.__name__} module...")

		data_path_list = glob(data_path_glob)
		if not data_path_list:
			raise FileNotFoundError(f"data does not exits in {data_path_glob}")

		assert file_type in [
			"pdf",
			"csv",
			"json",
			"md",
			"html",
			"xml",
			"all_files",
		], f"search type {file_type} is not supported"

		# extract only files from data_path_list based on the file_type set in the YAML file
		data_paths = (
			[
				data_path
				for data_path in data_path_list
				if os.path.basename(data_path).split(".")[-1] == file_type
			]
			if file_type != "all_files"
			else data_path_list
		)

		if func.__name__ == "langchain_parse":
			parse_method = parse_method.lower()
			if parse_method == "directory":
				path_split_list = data_path_glob.split("/")
				glob_path = path_split_list.pop()
				folder_path = "/".join(path_split_list)
				kwargs.update({"glob": glob_path, "path": folder_path})
				result = func(
					data_path_list=data_paths, parse_method=parse_method, **kwargs
				)
			else:
				result = func(
					data_path_list=data_paths, parse_method=parse_method, **kwargs
				)
		elif func.__name__ in ["clova_ocr", "llama_parse", "table_hybrid_parse"]:
			result = func(data_path_list=data_paths, **kwargs)
		else:
			raise ValueError(f"Unsupported module_type: {func.__name__}")
		result = _add_last_modified_datetime(result)
		return result

	return wrapper


def _add_last_modified_datetime(result):
	last_modified_datetime_lst = list(
		map(lambda x: get_file_metadata(x)["last_modified_datetime"], result[1])
	)
	result_with_dates = result + (last_modified_datetime_lst,)
	return result_with_dates



import abc

import pandas as pd

from autorag.nodes.retrieval.base import BaseRetrieval
from autorag.utils import result_to_dataframe
from autorag.utils.util import pop_params, fetch_contents


class HybridRetrieval(BaseRetrieval, metaclass=abc.ABCMeta):
	def __init__(self, project_dir: str, *args, **kwargs):
		super().__init__(project_dir)

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		previous_info = self.cast_to_run(previous_result, *args, **kwargs)
		_pure_params = pop_params(self._pure, kwargs)
		ids, scores = self._pure(previous_info, **_pure_params)
		contents = fetch_contents(self.corpus_df, ids)
		return contents, ids, scores

	def cast_to_run(self, previous_result: pd.DataFrame, *args, **kwargs):
		return hybrid_cast(previous_result)

	@classmethod
	def cast_to_run_class(cls, previous_result: pd.DataFrame):
		return hybrid_cast(previous_result)


def hybrid_cast(
	previous_result: pd.DataFrame,
):
	assert "query" in previous_result.columns, "previous_result must have query column."
	queries = previous_result["query"].tolist()

	assert "retrieved_contents_semantic" in previous_result.columns
	assert "retrieved_contents_lexical" in previous_result.columns
	assert "retrieve_scores_semantic" in previous_result.columns
	assert "retrieve_scores_lexical" in previous_result.columns
	assert "retrieved_ids_semantic" in previous_result.columns
	assert "retrieved_ids_lexical" in previous_result.columns

	contents_semantic = previous_result["retrieved_contents_semantic"].tolist()
	contents_lexical = previous_result["retrieved_contents_lexical"].tolist()
	scores_semantic = previous_result["retrieve_scores_semantic"].tolist()
	scores_lexical = previous_result["retrieve_scores_lexical"].tolist()
	ids_semantic = previous_result["retrieved_ids_semantic"].tolist()
	ids_lexical = previous_result["retrieved_ids_lexical"].tolist()

	return {
		"queries": queries,
		"retrieved_contents_semantic": contents_semantic,
		"retrieved_contents_lexical": contents_lexical,
		"retrieve_scores_semantic": scores_semantic,
		"retrieve_scores_lexical": scores_lexical,
		"retrieved_ids_semantic": ids_semantic,
		"retrieved_ids_lexical": ids_lexical,
	}



from typing import Dict, List

from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.base.llms.types import ChatMessage, MessageRole, ChatResponse
from llama_index.llms.openai.utils import to_openai_message_dicts
from openai import AsyncClient
from pydantic import BaseModel

from autorag.data.qa.filter.prompt import FILTER_PROMPT


class Response(BaseModel):
	is_passage_dependent: bool


async def passage_dependency_filter_openai(
	row: Dict,
	client: AsyncClient,
	model_name: str = "gpt-4o-mini-2024-07-18",
	lang: str = "en",
) -> bool:
	"""
	This will drop passage-dependent question rows.
	Passage-dependent questions are questions that the answer will change depending on what passage you choose.
	The passage-dependent questions will not be good for RAG evaluation, because any retrieval system can't find the right passage with passage-dependent question.
	For example, when someone asks "What is the highest score according to the table?" the answer will be different depending on the table.
	And what is the table? The retrieval system can't find the right passage with this question.
	You can use this filter with the ` batch_filter ` function at `QA` class.

	:param row: The row dict from QA dataset.
	    :param client: The OpenAI client.
	    :param model_name: The model name.
	            You have to use gpt-4o-2024-08-06 or gpt-4o-mini-2024-07-18.
	    :param lang: The supported language is en, ko or ja.
	:return: False if the row question is a passage-dependent question (to be filtered).
	"""
	assert "query" in row.keys(), "query column is not in the row."
	system_prompt: List[ChatMessage] = FILTER_PROMPT["passage_dependency"][lang]
	query = row["query"]
	completion = await client.beta.chat.completions.parse(
		model=model_name,
		messages=to_openai_message_dicts(
			system_prompt
			+ [
				ChatMessage(
					role=MessageRole.USER,
					content=f"Question: {query}\nIs this the question passage dependent?",
				)
			]
		),
		response_format=Response,
	)
	return not completion.choices[0].message.parsed.is_passage_dependent


async def passage_dependency_filter_llama_index(
	row: Dict,
	llm: BaseLLM,
	lang: str = "en",
) -> bool:
	"""
	This will drop passage-dependent question rows.
	Passage-dependent questions are questions that the answer will change depending on what passage you choose.
	The passage-dependent questions will not be good for RAG evaluation, because any retrieval system can't find the right passage with passage-dependent question.
	For example, when someone asks "What is the highest score according to the table?" the answer will be different depending on the table.
	And what is the table? The retrieval system can't find the right passage with this question.
	You can use this filter with the ` batch_filter ` function at `QA` class.

	:param row: The row dict from QA dataset.
	:param llm: The Llama index llm instance.
	            It will be good if you set max tokens to low for saving tokens.
	    :param lang: The supported language is en, ko or ja.
	:return: False if the row question is a passage-dependent question (to be filtered).
	"""
	assert "query" in row.keys(), "query column is not in the row."
	system_prompt: List[ChatMessage] = FILTER_PROMPT["passage_dependency"][lang]
	query = row["query"]
	response: ChatResponse = await llm.achat(
		messages=system_prompt
		+ [
			ChatMessage(
				role=MessageRole.USER,
				content=f"Question: {query}\nIs this the question passage dependent?",
			)
		]
	)
	result_str = response.message.content
	return "true" not in result_str.lower().strip()



from pathlib import Path
from typing import Any, List, Tuple

import numpy as np
import pandas as pd

from autorag.nodes.passagereranker.base import BasePassageReranker


from autorag.utils.util import (
	make_batch,
	sort_by_scores,
	flatten_apply,
	select_top_k,
	result_to_dataframe,
	pop_params,
	empty_cuda_cache,
)


class OpenVINOReranker(BasePassageReranker):
	def __init__(
		self,
		project_dir: str,
		model: str = "BAAI/bge-reranker-large",
		*args,
		**kwargs,
	):
		super().__init__(project_dir)

		try:
			from huggingface_hub import HfApi
			from transformers import AutoTokenizer
		except ImportError as e:
			raise ValueError(
				"Could not import huggingface_hub python package. "
				"Please install it with: "
				"`pip install -U huggingface_hub`."
			) from e

		def require_model_export(
			model_id: str, revision: Any = None, subfolder: Any = None
		) -> bool:
			model_dir = Path(model_id)
			if subfolder is not None:
				model_dir = model_dir / subfolder
			if model_dir.is_dir():
				return (
					not (model_dir / "openvino_model.xml").exists()
					or not (model_dir / "openvino_model.bin").exists()
				)
			hf_api = HfApi()
			try:
				model_info = hf_api.model_info(model_id, revision=revision or "main")
				normalized_subfolder = (
					None if subfolder is None else Path(subfolder).as_posix()
				)
				model_files = [
					file.rfilename
					for file in model_info.siblings
					if normalized_subfolder is None
					or file.rfilename.startswith(normalized_subfolder)
				]
				ov_model_path = (
					"openvino_model.xml"
					if subfolder is None
					else f"{normalized_subfolder}/openvino_model.xml"
				)
				return (
					ov_model_path not in model_files
					or ov_model_path.replace(".xml", ".bin") not in model_files
				)
			except Exception:
				return True

		try:
			from optimum.intel.openvino import OVModelForSequenceClassification
		except ImportError:
			raise ImportError(
				"Please install optimum package to use OpenVINOReranker"
				"pip install 'optimum[openvino,nncf]'"
			)

		model_kwargs = pop_params(
			OVModelForSequenceClassification.from_pretrained, kwargs
		)

		if require_model_export(model):
			# use remote model
			self.model = OVModelForSequenceClassification.from_pretrained(
				model, export=True, **model_kwargs
			)
		else:
			# use local model
			self.model = OVModelForSequenceClassification.from_pretrained(
				model, **model_kwargs
			)

		self.tokenizer = AutoTokenizer.from_pretrained(model)

	def __del__(self):
		del self.model
		del self.tokenizer
		empty_cuda_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.get("top_k", 3)
		batch = kwargs.get("batch", 64)
		return self._pure(queries, contents, ids, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents based on their relevance to a query using MonoT5.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved

		:param batch: The number of queries to be processed in a batch
		:return: tuple of lists containing the reranked contents, ids, and scores
		"""
		# Retrieve the tokens used by the model to represent false and true predictions

		nested_list = [
			list(map(lambda x: [query, x], content_list))
			for query, content_list in zip(queries, contents_list)
		]

		rerank_scores = flatten_apply(
			openvino_run_model,
			nested_list,
			model=self.model,
			batch_size=batch,
			tokenizer=self.tokenizer,
		)

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": rerank_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def openvino_run_model(
	input_texts,
	model,
	batch_size: int,
	tokenizer,
):
	batch_input_texts = make_batch(input_texts, batch_size)
	results = []
	for batch_texts in batch_input_texts:
		input_tensors = tokenizer(
			batch_texts,
			padding=True,
			truncation=True,
			return_tensors="pt",
		)

		outputs = model(**input_tensors, return_dict=True)
		if outputs[0].shape[1] > 1:
			scores = outputs[0][:, 1]
		else:
			scores = outputs[0].flatten()

		scores = list(map(float, (1 / (1 + np.exp(-np.array(scores))))))
		results.extend(scores)
	return results



import multiprocessing as mp
from itertools import chain
from typing import List, Tuple

from autorag.data import parse_modules
from autorag.data.parse.base import parser_node


@parser_node
def langchain_parse(
	data_path_list: List[str], parse_method: str, **kwargs
) -> Tuple[List[str], List[str], List[int]]:
	"""
	Parse documents to use langchain document_loaders(parse) method

	:param data_path_list: The list of data paths to parse.
	:param parse_method: A langchain document_loaders(parse) method to use.
	:param kwargs: The extra parameters for creating the langchain document_loaders(parse) instance.
	:return: tuple of lists containing the parsed texts, path and pages.
	"""
	if parse_method in ["directory", "unstructured"]:
		results = parse_all_files(data_path_list, parse_method, **kwargs)
		texts, path = results[0], results[1]
		pages = [-1] * len(texts)

	else:
		num_workers = mp.cpu_count()
		# Execute parallel processing
		with mp.Pool(num_workers) as pool:
			results = pool.starmap(
				langchain_parse_pure,
				[(data_path, parse_method, kwargs) for data_path in data_path_list],
			)

		texts, path, pages = (list(chain.from_iterable(item)) for item in zip(*results))

	return texts, path, pages


def langchain_parse_pure(
	data_path: str, parse_method: str, kwargs
) -> Tuple[List[str], List[str], List[int]]:
	"""
	Parses a single file using the specified parse method.

	Args:
	    data_path (str): The file path to parse.
	    parse_method (str): The parsing method to use.
	    kwargs (Dict): Additional keyword arguments for the parsing method.

	Returns:
	    Tuple[str, str]: A tuple containing the parsed text and the file path.
	"""

	parse_instance = parse_modules[parse_method](data_path, **kwargs)

	# Load the text from the file
	documents = parse_instance.load()

	texts = list(map(lambda x: x.page_content, documents))
	path = [data_path] * len(texts)
	if parse_method in ["pymupdf", "pdfplumber", "pypdf", "pypdfium2"]:
		pages = list(range(1, len(documents) + 1))
	else:
		pages = [-1] * len(texts)

	# Clean up the parse instance
	del parse_instance

	return texts, path, pages


def parse_all_files(
	data_path_list: List[str], parse_method: str, **kwargs
) -> Tuple[List[str], List[str]]:
	if parse_method == "unstructured":
		parse_instance = parse_modules[parse_method](data_path_list, **kwargs)
	elif parse_method == "directory":
		parse_instance = parse_modules[parse_method](**kwargs)
	else:
		raise ValueError(f"Unsupported parse method: {parse_method}")
	docs = parse_instance.load()
	texts = [doc.page_content for doc in docs]
	file_names = [doc.metadata["source"] for doc in docs]

	del parse_instance
	return texts, file_names



from llama_index.core.base.llms.types import ChatMessage, MessageRole

QUERY_GEN_PROMPT = {
	"factoid_single_hop": {
		"en": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""You're an AI tasked to convert Text into a factoid question.
Factoid questions are those seeking brief, factual information that can be easily verified. They typically require a yes or no answer or a brief explanation and often inquire about specific details such as dates, names, places, or events.

Examples of factoid questions include:

- What is the capital of France?
- Who invented the light bulb?
- When was Wikipedia founded?

Instructions:
1. Questions MUST BE extracted from given Text
2. Questions should be as detailed as possible from Text
3. Create questions that ask about factual information from the Text
4. Do not mention any of these in the questions: "in the given text", "in the provided information", etc.
Users do not know the passage source of the question, so it should not be mentioned in the question.
5. Do not ask about the file name or the file title. Ask about the content of the file.
For example, avoid to write questions like `What is the file name of the document?`""",
			)
		],
		"ko": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""당신은 주어진 Text를 '사실 질문'으로 변환하는 AI입니다.

사실 질문(factoid questions)이란 사실적인 정보를 요구하는 질문으로, 쉽게 검증할 수 있는 답변을 필요로 합니다. 일반적으로 예/아니오 답변이나 간단한 설명을 요구하며, 날짜, 이름, 장소 또는 사건과 같은 구체적인 세부사항에 대해 묻는 질문입니다.

사실 질문의 예는 다음과 같습니다:

	•	프랑스의 수도는 어디입니까?
	•	전구를 발명한 사람은 누구입니까?
	•	위키피디아는 언제 설립되었습니까?

지침:
	1.	질문은 반드시 주어진 Text를 기반으로 작성되어야 합니다.
	2.	질문은 Text를 기반으로 가능한 한 구체적으로 작성되어야 합니다.
	3.	Text에서 사실적 정보를 요구하는 질문을 만들어야 합니다. 즉, Text를 기반으로 사실 질문을 만드세요.
	4.	질문에 “주어진 Text에서” 또는 “제공된 단락에서”와 같은 표현을 포함해서는 안 됩니다.
사용자는 질문의 출처가 Text라는 것을 모르기 때문에 반드시 그 출처를 언급해서는 안 됩니다.
	5.	파일 이름이나 파일 제목에 대한 질문을 하지 마세요. 파일의 내용에 대해 물어보세요.
예를 들어, '문서의 파일 이름은 무엇입니까?'와 같은 질문을 작성하지 마세요.
	6.	질문을 한국어로 작성하세요.""",
			)
		],
		"ja": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""あなたは与えられたTextを「実は質問」に変換するAIです。

事実質問(factoid questions)とは、事実的な情報を求める質問であり、容易に検証できる回答を必要とします。 一般的に、「はい/いいえ」の返答や簡単な説明を要求し、日付、名前、場所、または事件のような具体的な詳細事項について尋ねる質問です。

事実質問の例は下記の通りです:

	• フランスの首都はどこですか？
	• 電球を発明したのは誰ですか？
	• ウィキペディアはいつ設立されましたか？

指針:
	1. 質問は、必ず与えられたTextに基づいて作成されなければなりません。
	2. 質問は、Textに基づいて可能な限り具体的に作成されなければなりません。
	3. Textで事実的な情報を要求する質問を作らなければなりません。 つまり、Textに基づいて事実質問を作成してください。
	4. 質問に「与えられたTextで」または「提供された段落で」のような表現を含めてはいけません。
ユーザーは質問の出所がTextだということを知らないので、絶対にその出所を言及してはいけません。
	5. ファイル名やファイルのタイトルを訊かないでください。代わりに、ファイルの内容について聞いてください。
例えば、「このドキュメントのファイル名は何ですか？」や「ファイル名は何ですか？」というような質問は書かないでください。
	6. 質問を日本語で作成してください。""",
			)
		],
	},
	"concept_completion": {
		"en": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""You're an AI tasked to convert Text into a "Concept Completion" Question.
A “concept completion” question asks directly about the essence or identity of a concept.

Follow the following instructions.
Instructions:
1. Questions MUST BE extracted from given Text
2. Questions should be as detailed as possible from Text
3. Create questions that ask about information from the Text
4. MUST include specific keywords from the Text.
5. Do not mention any of these in the questions: "in the given text", "in the provided information", etc.
Users do not know the passage source of the question, so it should not be mentioned in the question.
6. Do not ask about the file name or the file title. Ask about the content of the file.
For example, avoid to write questions like `What is the file name of the document?""",
			)
		],
		"ko": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""당신은 Text를 “개념 완성” 질문으로 변환하는 AI입니다.
"개념 완성" 질문은 개념의 본질이나 정체성에 대해 직접적으로 묻는 질문입니다.

다음 지시사항을 따르세요.
지시사항:
1.	질문은 반드시 주어진 Text를 기반으로 작성되어야 합니다.
2.	질문은 Text를 기반으로 가능한 한 자세하게 작성되어야 합니다.
3.	Text에서 제공된 정보를 묻는 질문을 생성하세요.
4.	Text의 특정 키워드를 반드시 질문에 포함하세요.
5.	질문에 “주어진 Text에서” 또는 “제공된 단락에서”와 같은 표현을 포함해서는 안 됩니다.
사용자는 질문의 출처가 Text라는 것을 모르기 때문에 반드시 그 출처를 언급해서는 안 됩니다.
6.	파일 이름이나 파일 제목에 대한 질문을 하지 마세요. 파일의 내용에 대해 물어보세요.
예를 들어, '문서의 파일 이름은 무엇입니까?'와 같은 질문을 작성하지 마세요.
7.	질문을 한국어로 작성하세요.""",
			)
		],
		"ja": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="""あなたはTextを「概念完成」の質問に変換するAIです。
「概念完成」の質問は概念の本質やアイデンティティについて直接に尋ねる質問です。

次の指示に従ってください。
指示事項:
1. 質問は、必ず与えられたTextに基づいて作成されなければなりません。
2. 質問は、Textに基づいてできるだけ詳しく作成されなければなりません。
3. Textで提供された情報を尋ねる質問を作成してください。
4. Textの特定のキーワードを必ず質問に含めてください。
5. 質問に「与えられたTextで」または「提供された段落で」のような表現を含めてはいけません。
ユーザーは質問の出所がTextだということを知らないので、絶対にその出所を言及してはいけません。
6. ファイル名やファイルのタイトルを訊かないでください。代わりに、ファイルの内容について聞いてください。
例えば、「このドキュメントのファイル名は何ですか？」や「ファイル名は何ですか？」というような質問は書かないでください。
7. 質問を日本語で作成してください。""",
			)
		],
	},
	"two_hop_incremental": {
		"en": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="Generate a multi-hop question for the given answer which requires reference to all of the given documents.",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Document 1: The Municipality of Nuevo Laredo is located in the Mexican state of Tamaulipas.
Document 2: The Ciudad Deportiva (Sports City ¨ ¨) is a sports
complex in Nuevo Laredo, Mexico. It is home to the Tecolotes de
Nuevo Laredo Mexican Baseball League team and ...""",
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="""Answer: Tamaulipas
One-hop question (using Document 1): In which Mexican state is Nuevo Laredo located?
Two-hop question (using Document 2):  In which Mexican state can one find the Ciudad Deportiva, home to the Tecolotes de Nuevo Laredo?""",
			),
		],
		"ko": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="Generate a multi-hop question for the given answer which requires reference to all of the given documents.",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Document 1: The Municipality of Nuevo Laredo is located in the Mexican state of Tamaulipas.
Document 2: The Ciudad Deportiva (Sports City ¨ ¨) is a sports
complex in Nuevo Laredo, Mexico. It is home to the Tecolotes de
Nuevo Laredo Mexican Baseball League team and ...""",
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="""Answer: Tamaulipas
One-hop question (using Document 1): In which Mexican state is Nuevo Laredo located?
Two-hop question (using Document 2):  In which Mexican state can one find the Ciudad Deportiva, home to the Tecolotes de Nuevo Laredo?""",
			),
		],
		"ja": [
			ChatMessage(
				role=MessageRole.SYSTEM,
				content="与えられた答えに対するマルチホップ質問を生成し、与えられたすべての文書を参照する必要があります。",
			),
			ChatMessage(
				role=MessageRole.USER,
				content="""Document 1: ヌエヴォ·ラレド自治体はメキシコのタマウリパス州にあります。
シウダー・デポルティーバ（スポーツ・シティ）は、
メキシコのヌエボ・ラレドにある複合スポーツ施設です。""",
			),
			ChatMessage(
				role=MessageRole.ASSISTANT,
				content="""Answer: Tamaulipas
One-hop question (using Document 1): ヌエヴォ·ラレド自治体はどのメキシコの州にありますか？
Two-hop question (using Document 2): ヌエヴォ·ラレドのテコロテス·デ·テコロテスの故郷であるメキシコの州はどこですか？""",
			),
		],
	},
}

# Experimental feature
QUERY_GEN_PROMPT_EXTRA = {
	"multiple_queries": {
		"en": "\nAdditional instructions:\n  - Please make {n} questions.",
		"ko": "\n추가 지침:\n  - 질문은 {n}개를 만드세요.",
		"ja": "\n追加指示:\n  - 質問を{n}個作成してください。",
	}
}



import itertools
from typing import Dict

from openai import AsyncClient
from pydantic import BaseModel

from autorag.data.qa.generation_gt.base import add_gen_gt
from autorag.data.qa.generation_gt.prompt import GEN_GT_SYSTEM_PROMPT


class Response(BaseModel):
	answer: str


async def make_gen_gt_openai(
	row: Dict,
	client: AsyncClient,
	system_prompt: str,
	model_name: str = "gpt-4o-2024-08-06",
):
	retrieval_gt_contents = list(
		itertools.chain.from_iterable(row["retrieval_gt_contents"])
	)
	query = row["query"]
	passage_str = "\n".join(retrieval_gt_contents)
	user_prompt = f"Text:\n<|text_start|>\n{passage_str}\n<|text_end|>\n\nQuestion:\n{query}\n\nAnswer:"

	completion = await client.beta.chat.completions.parse(
		model=model_name,
		messages=[
			{"role": "system", "content": system_prompt},
			{"role": "user", "content": user_prompt},
		],
		temperature=0.0,
		response_format=Response,
	)
	response: Response = completion.choices[0].message.parsed
	return add_gen_gt(row, response.answer)


async def make_concise_gen_gt(
	row: Dict,
	client: AsyncClient,
	model_name: str = "gpt-4o-2024-08-06",
	lang: str = "en",
):
	"""
	Generate concise generation_gt using OpenAI Structured Output for preventing errors.
	It generates a concise answer, so it is generally a word or just a phrase.

	:param row: The input row of the qa dataframe.
	:param client: The OpenAI async client.
	:param model_name: The model name that supports structured output.
	    It has to be "gpt-4o-2024-08-06" or "gpt-4o-mini-2024-07-18".
	:param lang: The language code of the prompt.
		Default is "en".
	:return: The output row of the qa dataframe with added "generation_gt" in it.
	"""
	return await make_gen_gt_openai(
		row, client, GEN_GT_SYSTEM_PROMPT["concise"][lang], model_name
	)


async def make_basic_gen_gt(
	row: Dict,
	client: AsyncClient,
	model_name: str = "gpt-4o-2024-08-06",
	lang: str = "en",
):
	"""
	Generate basic generation_gt using OpenAI Structured Output for preventing errors.
	It generates a "basic" answer, and its prompt is simple.

	:param row: The input row of the qa dataframe.
	:param client: The OpenAI async client.
	:param model_name: The model name that supports structured output.
	    It has to be "gpt-4o-2024-08-06" or "gpt-4o-mini-2024-07-18".
	:param lang: The language code of the prompt.
		Default is "en".
	:return: The output row of the qa dataframe with added "generation_gt" in it.
	"""
	return await make_gen_gt_openai(
		row, client, GEN_GT_SYSTEM_PROMPT["basic"][lang], model_name
	)



from typing import Optional, Dict

from autorag.support import get_support_modules


def make_generator_callable_param(generator_dict: Optional[Dict]):
	if "generator_module_type" not in generator_dict.keys():
		generator_dict = {
			"generator_module_type": "llama_index_llm",
			"llm": "openai",
			"model": "gpt-4o-mini",
		}
	module_str = generator_dict.pop("generator_module_type")
	module_class = get_support_modules(module_str)
	module_param = generator_dict
	return module_class, module_param



from .long_context_reorder import LongContextReorder
from .window_replacement import WindowReplacement
from .fstring import Fstring
from .chat_fstring import ChatFstring


__all__ = [
	"Fstring",
	"ChatFstring",
	"LongContextReorder",
	"WindowReplacement",
]



from copy import deepcopy
from typing import Union, List, Dict, Tuple, Any

from autorag.embedding.base import EmbeddingModel


def cast_metrics(
	metrics: Union[List[str], List[Dict]],
) -> Tuple[List[str], List[Dict[str, Any]]]:
	"""
	 Turn metrics to list of metric names and parameter list.

	:param metrics: List of string or dictionary.
	:return: The list of metric names and dictionary list of metric parameters.
	"""
	metrics_copy = deepcopy(metrics)
	if not isinstance(metrics_copy, list):
		raise ValueError("metrics must be a list of string or dictionary.")
	if isinstance(metrics_copy[0], str):
		return metrics_copy, [{} for _ in metrics_copy]
	elif isinstance(metrics_copy[0], dict):
		# pop 'metric_name' key from dictionary
		metric_names = list(map(lambda x: x.pop("metric_name"), metrics_copy))
		metric_params = [
			dict(
				map(
					lambda x, y: cast_embedding_model(x, y),
					metric.keys(),
					metric.values(),
				)
			)
			for metric in metrics_copy
		]
		return metric_names, metric_params
	else:
		raise ValueError("metrics must be a list of string or dictionary.")


def cast_embedding_model(key, value):
	if key == "embedding_model":
		return key, EmbeddingModel.load(value)()
	else:
		return key, value



import os
import pathlib
from copy import deepcopy
from typing import List, Dict, Optional, Union

import pandas as pd
import tokenlog

from autorag.evaluation import evaluate_generation
from autorag.evaluation.util import cast_metrics
from autorag.schema.metricinput import MetricInput
from autorag.strategy import measure_speed, filter_by_threshold, select_best
from autorag.support import get_support_modules
from autorag.utils import validate_qa_dataset
from autorag.utils.util import make_combinations, explode, split_dataframe


def run_prompt_maker_node(
	modules: List,
	module_params: List[Dict],
	previous_result: pd.DataFrame,
	node_line_dir: str,
	strategies: Dict,
) -> pd.DataFrame:
	"""
	Run prompt maker node.
	With this function, you can select the best prompt maker module.
	As default, when you can use only one module, the evaluation will be skipped.
	If you want to select the best prompt among modules, you can use strategies.
	When you use them, you must pass 'generator_modules' and its parameters at strategies.
	Because it uses generator modules and generator metrics for evaluation this module.
	It is recommended to use one params and modules for evaluation,
	but you can use multiple params and modules for evaluation.
	When you don't set generator module at strategies, it will use the default generator module.
	The default generator module is llama_index_llm with openai gpt-3.5-turbo model.

	:param modules: Prompt maker module classes to run.
	:param module_params: Prompt maker module parameters.
	:param previous_result: Previous result dataframe.
	    Could be query expansion's best result or qa data.
	:param node_line_dir: This node line's directory.
	:param strategies: Strategies for prompt maker node.
	:return: The best result dataframe.
	    It contains previous result columns and prompt maker's result columns which is 'prompts'.
	"""
	if not os.path.exists(node_line_dir):
		os.makedirs(node_line_dir)
	node_dir = os.path.join(node_line_dir, "prompt_maker")
	if not os.path.exists(node_dir):
		os.makedirs(node_dir)
	project_dir = pathlib.PurePath(node_line_dir).parent.parent

	# run modules
	results, execution_times = zip(
		*map(
			lambda task: measure_speed(
				task[0].run_evaluator,
				project_dir=project_dir,
				previous_result=previous_result,
				**task[1],
			),
			zip(modules, module_params),
		)
	)
	average_times = list(map(lambda x: x / len(results[0]), execution_times))

	# get average token usage
	token_usages = []
	for i, result in enumerate(results):
		token_logger = tokenlog.getLogger(
			f"prompt_maker_{i}", strategies.get("tokenizer", "openai/gpt-oss-20b")
		)
		token_logger.query_batch(result["prompts"].tolist())
		token_usages.append(token_logger.get_token_usage() / len(result))

	# save results to folder
	filepaths = list(
		map(lambda x: os.path.join(node_dir, f"{x}.parquet"), range(len(modules)))
	)
	list(
		map(lambda x: x[0].to_parquet(x[1], index=False), zip(results, filepaths))
	)  # execute save to parquet
	filenames = list(map(lambda x: os.path.basename(x), filepaths))

	# make summary file
	summary_df = pd.DataFrame(
		{
			"filename": filenames,
			"module_name": list(map(lambda module: module.__name__, modules)),
			"module_params": module_params,
			"execution_time": average_times,
			"average_prompt_token": token_usages,
		}
	)

	metric_names, metric_params = cast_metrics(strategies.get("metrics"))

	# Run evaluation when there are more than one module.
	if len(modules) > 1:
		# pop general keys from strategies (e.g. metrics, speed_threshold)
		general_key = ["metrics", "speed_threshold", "token_threshold", "tokenizer"]
		general_strategy = dict(
			filter(lambda x: x[0] in general_key, strategies.items())
		)
		extra_strategy = dict(
			filter(lambda x: x[0] not in general_key, strategies.items())
		)

		# first, filter by threshold if it is enabled.
		if general_strategy.get("speed_threshold") is not None:
			results, filenames = filter_by_threshold(
				results, average_times, general_strategy["speed_threshold"], filenames
			)

		# Calculate tokens and save to summary
		if general_strategy.get("token_threshold") is not None:
			results, filenames = filter_by_threshold(
				results, token_usages, general_strategy["token_threshold"], filenames
			)

		# run metrics before filtering
		if metric_names is None or len(metric_names) <= 0:
			raise ValueError(
				"You must at least one metrics for prompt maker evaluation."
			)

		# get generator modules from strategy
		generator_callables, generator_params = make_generator_callable_params(
			extra_strategy
		)

		# get generation_gt
		qa_data = pd.read_parquet(
			os.path.join(project_dir, "data", "qa.parquet"), engine="pyarrow"
		)
		validate_qa_dataset(qa_data)
		generation_gt = qa_data["generation_gt"].tolist()
		generation_gt = list(map(lambda x: x.tolist(), generation_gt))

		metric_inputs = [MetricInput(generation_gt=gen_gt) for gen_gt in generation_gt]

		all_prompts = []
		for result in results:
			all_prompts.extend(result["prompts"].tolist())

		evaluation_result_all = evaluate_one_prompt_maker_node(
			all_prompts,
			generator_callables,
			generator_params,
			metric_inputs * len(results),
			general_strategy["metrics"],
			project_dir,
			strategy_name=strategies.get("strategy", "mean"),
		)
		evaluation_results = split_dataframe(
			evaluation_result_all, chunk_size=len(results[0])
		)

		evaluation_df = pd.DataFrame(
			{
				"filename": filenames,
				**{
					f"prompt_maker_{metric_name}": list(
						map(lambda x: x[metric_name].mean(), evaluation_results)
					)
					for metric_name in metric_names
				},
			}
		)
		summary_df = pd.merge(
			on="filename", left=summary_df, right=evaluation_df, how="left"
		)

		best_result, best_filename = select_best(
			evaluation_results,
			metric_names,
			filenames,
			strategies.get("strategy", "mean"),
		)
		# change metric name columns to prompt_maker_metric_name
		best_result = best_result.rename(
			columns={
				metric_name: f"prompt_maker_{metric_name}"
				for metric_name in metric_names
			}
		)
		best_result = best_result.drop(columns=["generated_texts"])
	else:
		best_result, best_filename = results[0], filenames[0]

	# add 'is_best' column at summary file
	summary_df["is_best"] = summary_df["filename"] == best_filename

	best_result = pd.concat([previous_result, best_result], axis=1)

	# save files
	summary_df.to_csv(os.path.join(node_dir, "summary.csv"), index=False)
	best_result.to_parquet(
		os.path.join(node_dir, f"best_{os.path.splitext(best_filename)[0]}.parquet"),
		index=False,
	)

	return best_result


def make_generator_callable_params(strategy_dict: Dict):
	node_dict = deepcopy(strategy_dict)
	generator_module_list: Optional[List[Dict]] = node_dict.pop(
		"generator_modules", None
	)
	if generator_module_list is None:
		generator_module_list = [
			{
				"module_type": "llama_index_llm",
				"llm": "openai",
				"model": "gpt-3.5-turbo",
			}
		]
	node_params = node_dict
	modules = list(
		map(
			lambda module_dict: get_support_modules(module_dict.pop("module_type")),
			generator_module_list,
		)
	)
	param_combinations = list(
		map(
			lambda module_dict: make_combinations({**module_dict, **node_params}),
			generator_module_list,
		)
	)
	return explode(modules, param_combinations)


def evaluate_one_prompt_maker_node(
	prompts: List[str],
	generator_classes: List,
	generator_params: List[Dict],
	metric_inputs: List[MetricInput],
	metrics: Union[List[str], List[Dict]],
	project_dir,
	strategy_name: str,
) -> pd.DataFrame:
	input_df = pd.DataFrame({"prompts": prompts})
	generator_results = list(
		map(
			lambda x: x[0].run_evaluator(
				project_dir=project_dir, previous_result=input_df, **x[1]
			),
			zip(generator_classes, generator_params),
		)
	)
	evaluation_results = list(
		map(
			lambda x: evaluate_generator_result(x[0], metric_inputs, metrics),
			zip(generator_results, generator_classes),
		)
	)
	metric_names = (
		list(map(lambda x: x["metric_name"], metrics))
		if isinstance(metrics[0], dict)
		else metrics
	)
	best_result, _ = select_best(
		evaluation_results, metric_names, strategy_name=strategy_name
	)
	best_result = pd.concat([input_df, best_result], axis=1)
	return best_result  # it has 'generated_texts' column


def evaluate_generator_result(
	result_df: pd.DataFrame,
	metric_inputs: List[MetricInput],
	metrics: Union[List[str], List[Dict]],
) -> pd.DataFrame:
	@evaluate_generation(metric_inputs=metric_inputs, metrics=metrics)
	def evaluate(df):
		return df["generated_texts"].tolist()

	return evaluate(result_df)



import uuid
from typing import List, Optional

import pandas as pd
from llama_index.core import Document
from llama_index.core.schema import TextNode

from autorag.data.utils.util import (
	add_essential_metadata,
	add_essential_metadata_llama_text_node,
)
from autorag.utils.util import save_parquet_safe


def llama_documents_to_parquet(
	llama_documents: List[Document],
	output_filepath: Optional[str] = None,
	upsert: bool = False,
) -> pd.DataFrame:
	"""
	Llama Index documents to corpus dataframe.
	Corpus dataframe will be saved to filepath(file_dir/filename) if given.
	Return corpus dataframe whether the filepath is given.
	You can use this method to create corpus.parquet after load and chunk using Llama Index.

	:param llama_documents: List[Document]
	:param output_filepath: Optional filepath to save the parquet file.
	    If None, the function will return the processed_data as pd.DataFrame, but do not save as parquet.
	    File directory must exist. File extension must be .parquet
	:param upsert: If true, the function will overwrite the existing file if it exists.
	    Default is False.
	:return: Corpus data as pd.DataFrame
	"""

	doc_lst = pd.DataFrame(
		list(
			map(
				lambda doc: {
					"doc_id": str(uuid.uuid4()),
					"contents": doc.text,
					"metadata": add_essential_metadata(doc.metadata),
				},
				llama_documents,
			)
		)
	)

	processed_df = pd.DataFrame(doc_lst)

	if output_filepath is not None:
		save_parquet_safe(processed_df, output_filepath, upsert=upsert)

	return processed_df


def llama_text_node_to_parquet(
	text_nodes: List[TextNode],
	output_filepath: Optional[str] = None,
	upsert: bool = False,
) -> pd.DataFrame:
	"""
	Llama Index text nodes to corpus dataframe.
	Corpus dataframe will be saved to filepath(file_dir/filename) if given.
	Return corpus dataframe whether the filepath is given.
	You can use this method to create corpus.parquet after load and chunk using Llama Index.

	:param text_nodes: List of llama index text nodes.
	:param output_filepath: Optional filepath to save the parquet file.
	    If None, the function will return the processed_data as pd.DataFrame, but do not save as parquet.
	    File directory must exist. File extension must be .parquet
	:param upsert: If true, the function will overwrite the existing file if it exists.
	    Default is False.
	:return: Corpus data as pd.DataFrame
	"""
	corpus_df = pd.DataFrame(
		list(
			map(
				lambda node: {
					"doc_id": node.node_id,
					"contents": node.text,
					"metadata": add_essential_metadata_llama_text_node(
						node.metadata, node.relationships
					),
				},
				text_nodes,
			)
		)
	)

	if output_filepath is not None:
		save_parquet_safe(corpus_df, output_filepath, upsert=upsert)

	return corpus_df
